# Copyright 2018 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

from oslo_upgradecheck import upgradecheck
from sqlalchemy.schema import MetaData
from sqlalchemy.schema import Table

from designate.cmd import status
from designate.storage import sql
import designate.tests.functional


class TestDuplicateServiceStatus(designate.tests.functional.TestCase):
    def setUp(self):
        super().setUp()
        self.meta = MetaData()
        self.meta.bind = sql.get_read_engine()
        self.service_statuses_table = Table(
            'service_statuses', self.meta,
            autoload_with=sql.get_read_engine()
        )

    def test_success(self):
        fake_record = {'id': '1',
                       'service_name': 'worker',
                       'hostname': '203.0.113.1',
                       'status': 'UP',
                       'stats': '',
                       'capabilities': '',
                       }
        with sql.get_write_session() as session:
            query = (
                self.service_statuses_table.insert().
                values(fake_record)
            )
            session.execute(query)

            # Different hostname should be fine
            fake_record['id'] = '2'
            fake_record['hostname'] = 'otherhost'
            query = (
                self.service_statuses_table.insert().
                values(fake_record)
            )
            session.execute(query)

            # Different service_name should be fine
            fake_record['id'] = '3'
            fake_record['service_name'] = 'producer'
            query = (
                self.service_statuses_table.insert().
                values(fake_record)
            )
            session.execute(query)

            checks = status.Checks()
            self.assertEqual(upgradecheck.Code.SUCCESS,
                             checks._duplicate_service_status().code)

    @mock.patch('designate.storage.sql.get_read_session')
    @mock.patch('designate.storage.sql.get_read_engine')
    def test_failure(self, mock_get_engine, mock_get_read):
        mock_sql_execute = mock.Mock()
        mock_sql_fetchall = mock.Mock()

        mock_get_read().__enter__.return_value = mock_sql_execute
        mock_sql_execute.execute.return_value = mock_sql_fetchall
        mock_sql_fetchall.fetchall.return_value = [(2,)]

        checks = status.Checks()

        result = checks._duplicate_service_status().code
        self.assertEqual(upgradecheck.Code.FAILURE, result)



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import copy
from operator import attrgetter
from unittest import mock

from oslo_log import log as logging
from oslo_serialization import jsonutils
import oslotest.base
import testtools

from designate import exceptions
from designate import objects
from designate.objects import base
from designate.objects import fields

LOG = logging.getLogger(__name__)


@base.DesignateRegistry.register
class TestObject(objects.DesignateObject):
    fields = {
        'id': fields.AnyField(nullable=True),
        'name': fields.AnyField(nullable=True),
        'nested': fields.ObjectFields('TestObject', nullable=True),
        'nested_list': fields.ObjectFields('TestObjectList', nullable=True),
    }

    STRING_KEYS = [
        'id', 'name'
    ]


@base.DesignateRegistry.register
class TestObjectDict(TestObject, objects.DictObjectMixin):
    pass


@base.DesignateRegistry.register
class TestObjectList(objects.ListObjectMixin, objects.DesignateObject):
    LIST_ITEM_TYPE = TestObject

    fields = {
        'objects': fields.ListOfObjectsField('TestObject'),
    }


@base.DesignateRegistry.register
class TestValidatableObject(objects.DesignateObject):
    fields = {
        'id': fields.UUIDFields(),
        'nested': fields.ObjectFields('TestValidatableObject',
                                      nullable=True),
    }


class DesignateObjectTest(oslotest.base.BaseTestCase):
    def test_obj_to_repr(self):
        obj = TestObject.from_dict({
            'id': 1, 'name': 'example'
        })
        self.assertEqual(
            "<TestObject id:'1' name:'example'>",
            repr(obj)
        )

    def test_obj_to_str(self):
        obj = TestObject.from_dict({
            'id': 1, 'name': 'example'
        })
        self.assertEqual(
            "<TestObject id:'1' name:'example'>", str(obj)
        )

    def test_empty_obj_to_str(self):
        self.assertEqual(
            "<TestObject id:'None' name:'None'>", str(TestObject())
        )

    def test_record_to_str(self):
        obj = objects.Record.from_dict({
            'id': 1, 'recordset_id': '2', 'data': 'example'
        })
        self.assertEqual(
            "<Record id:'1' recordset_id:'2' data:'example'>", str(obj)
        )

    def test_obj_cls_from_name(self):
        cls = objects.DesignateObject.obj_cls_from_name('TestObject')
        self.assertEqual(TestObject, cls)

        cls = objects.DesignateObject.obj_cls_from_name('TestObjectDict')
        self.assertEqual(TestObjectDict, cls)

        cls = objects.DesignateObject.obj_cls_from_name('TestObjectList')
        self.assertEqual(TestObjectList, cls)

    def test_from_primitive(self):
        primitive = {
            'designate_object.name': 'TestObject',
            'designate_object.data': {
                'id': 1,
            },
            'designate_object.changes': [],
            'designate_object.namespace': 'designate',
            'designate_object.version': '1.0',
        }

        obj = objects.DesignateObject.from_primitive(primitive)

        # Validate it has been thawed correctly
        self.assertEqual(1, obj.id)

        # Ensure the ID field has a value
        self.assertTrue(obj.obj_attr_is_set('id'))

        # Ensure the name field has no value
        self.assertFalse(obj.obj_attr_is_set('name'))

        # Ensure the changes list is empty
        self.assertEqual(0, len(obj.obj_what_changed()))

    def test_from_primitive_recursive(self):
        primitive = {
            'designate_object.name': 'TestObject',
            'designate_object.data': {
                'id': 1,
                'nested': {
                    'designate_object.name': 'TestObject',
                    'designate_object.data': {
                        'id': 2,
                    },
                    'designate_object.changes': [],
                    'designate_object.namespace': 'designate',
                    'designate_object.version': '1.0',
                }
            },
            'designate_object.changes': [],
            'designate_object.namespace': 'designate',
            'designate_object.version': '1.0',
        }

        obj = objects.DesignateObject.from_primitive(primitive)

        # Validate it has been thawed correctly
        self.assertEqual(1, obj.id)
        self.assertEqual(2, obj.nested.id)

    def test_from_dict(self):
        obj = TestObject.from_dict({
            'id': 1,
        })

        # Validate it has been thawed correctly
        self.assertEqual(1, obj.id)

        # Ensure the ID field has a value
        self.assertTrue(obj.obj_attr_is_set('id'))

        # Ensure the name field has no value
        self.assertFalse(obj.obj_attr_is_set('name'))

        # Ensure the changes list has one entry for the id field
        self.assertEqual({'id'}, obj.obj_what_changed())

    def test_from_dict_recursive(self):
        obj = TestObject.from_dict({
            'id': 1,
            'nested': {
                'id': 2,
            },
        })

        # Validate it has been thawed correctly
        self.assertEqual(1, obj.id)
        self.assertEqual(2, obj.nested.id)

        # Ensure the changes list has two entries, one for the id field and the
        # other for the nested field
        self.assertEqual({'id', 'nested'}, obj.obj_what_changed())

        # Ensure the changes list has one entry for the id field
        self.assertEqual({'id'}, obj.nested.obj_what_changed())

    def test_from_dict_nested_list(self):
        obj = TestObject.from_dict({
            'id': 1,
            'nested_list': [{
                'id': 2,
            }, {
                'id': 3,
            }],
        })

        # Validate it has been thawed correctly
        self.assertEqual(1, obj.id)
        self.assertEqual(2, obj.nested_list[0].id)
        self.assertEqual(3, obj.nested_list[1].id)

        # Ensure the changes list has two entries, one for the id field and the
        # other for the nested field
        self.assertEqual({'id', 'nested_list'}, obj.obj_what_changed())

    def test_from_list(self):
        with testtools.ExpectedException(NotImplementedError):
            TestObject.from_list([])

    def test_init_invalid(self):
        with testtools.ExpectedException(TypeError):
            TestObject(extra_field='Fail')

    def test_hasattr(self):
        obj = TestObject()

        # Success Cases
        self.assertTrue(hasattr(obj, 'id'),
                        "Should have id attribute")
        self.assertTrue(hasattr(obj, 'name'),
                        "Should have name attribute")

        # Failure Cases
        self.assertFalse(hasattr(obj, 'email'),
                         "Should not have email attribute")
        self.assertFalse(hasattr(obj, 'names'),
                         "Should not have names attribute")

    def test_setattr(self):
        obj = TestObject()

        obj.id = 1
        self.assertEqual(1, obj.id)
        self.assertEqual(1, len(obj.obj_what_changed()))

        obj.name = 'MyName'
        self.assertEqual('MyName', obj.name)
        self.assertEqual(2, len(obj.obj_what_changed()))

    def test_setattr_neg(self):
        obj = TestObject()

        with testtools.ExpectedException(AttributeError):
            obj.badthing = 'demons'

    def test_to_primitive(self):
        obj = TestObject(id=1)

        # Ensure only the id attribute is returned
        primitive = obj.to_primitive()
        expected = {
            'designate_object.name': 'TestObject',
            'designate_object.data': {
                'id': 1,
            },
            'designate_object.changes': ['id'],
            'designate_object.namespace': 'designate',
            'designate_object.version': '1.0',
        }
        self.assertEqual(expected, primitive)

        # Set the name attribute to a None value
        obj.name = None

        # Ensure both the id and name attributes are returned
        primitive = obj.to_primitive()
        expected = {
            'designate_object.name': 'TestObject',
            'designate_object.data': {
                'id': 1,
                'name': None,
            },
            'designate_object.changes': ['id', 'name'],
            'designate_object.namespace': 'designate',
            'designate_object.version': '1.0',
        }
        self.assertEqual(expected, primitive)

    def test_to_primitive_recursive(self):
        obj = TestObject(id=1, nested=TestObject(id=2))

        # Ensure only the id attribute is returned
        primitive = obj.to_primitive()
        expected = {
            'designate_object.name': 'TestObject',
            'designate_object.data': {
                'id': 1,
                'nested': {
                    'designate_object.name': 'TestObject',
                    'designate_object.data': {
                        'id': 2,
                    },
                    'designate_object.changes': ['id'],
                    'designate_object.namespace': 'designate',
                    'designate_object.version': '1.0',
                }
            },
            'designate_object.changes': ['id', 'nested'],
            'designate_object.namespace': 'designate',
            'designate_object.version': '1.0',
        }
        self.assertEqual(expected, primitive)

    def test_to_dict(self):
        obj = TestObject(id=1)

        # Ensure only the id attribute is returned
        dict_ = obj.to_dict()
        expected = {
            'id': 1,
        }
        self.assertEqual(expected, dict_)

        # Set the name attribute to a None value
        obj.name = None

        # Ensure both the id and name attributes are returned
        dict_ = obj.to_dict()
        expected = {
            'id': 1,
            'name': None,
        }
        self.assertEqual(expected, dict_)

    def test_to_dict_recursive(self):
        obj = TestObject(id=1, nested=TestObject(id=2))

        # Ensure only the id attribute is returned
        dict_ = obj.to_dict()
        expected = {
            'id': 1,
            'nested': {
                'id': 2,
            },
        }

        self.assertEqual(expected, dict_)

    def test_update(self):
        obj = TestObject(id=1, name='test')
        obj.update({'id': 'new_id', 'name': 'new_name'})
        self.assertEqual('new_id', obj.id)
        self.assertEqual('new_name', obj.name)

    def test_update_unexpected_attribute(self):
        obj = TestObject(id=1, name='test')
        with testtools.ExpectedException(AttributeError):
            obj.update({'id': 'new_id', 'new_key': 3})

    def test_validate(self):
        obj = TestValidatableObject()

        # ID is required, so the object is not valid
        with testtools.ExpectedException(exceptions.InvalidObject):
            obj.validate()

        with testtools.ExpectedException(ValueError):
            obj.id = 'MyID'

        # Set the ID field to a valid value
        obj.id = 'ffded5c4-e4f6-4e02-a175-48e13c5c12a0'
        obj.validate()

    def test_validate_recursive(self):
        with testtools.ExpectedException(ValueError):
            TestValidatableObject(
                id='MyID',
                nested=TestValidatableObject(id='MyID'))

        with testtools.ExpectedException(ValueError):
            TestValidatableObject(
                id='ffded5c4-e4f6-4e02-a175-48e13c5c12a0',
                nested=TestValidatableObject(
                    id='MyID'))

        obj = TestValidatableObject(
            id='ffded5c4-e4f6-4e02-a175-48e13c5c12a0',
            nested=TestValidatableObject(
                id='ffded5c4-e4f6-4e02-a175-48e13c5c12a0'))
        obj.validate()

    def test_obj_attr_is_set(self):
        obj = TestObject()

        self.assertFalse(obj.obj_attr_is_set('name'))

        obj.name = "My Name"

        self.assertTrue(obj.obj_attr_is_set('name'))

    def test_obj_what_changed(self):
        obj = TestObject()

        self.assertEqual(set(), obj.obj_what_changed())

        obj.name = "My Name"

        self.assertEqual({'name'}, obj.obj_what_changed())

    def test_obj_get_changes(self):
        obj = TestObject()

        self.assertEqual({}, obj.obj_get_changes())

        obj.name = "My Name"

        self.assertEqual({'name': "My Name"}, obj.obj_get_changes())

    def test_obj_reset_changes(self):
        obj = TestObject()
        obj.name = "My Name"

        self.assertEqual(1, len(obj.obj_what_changed()))

        obj.obj_reset_changes()

        self.assertEqual(0, len(obj.obj_what_changed()))

    def test_obj_reset_changes_subset(self):
        obj = TestObject()
        obj.id = "My ID"
        obj.name = "My Name"

        self.assertEqual(2, len(obj.obj_what_changed()))

        obj.obj_reset_changes(['id'])

        self.assertEqual(1, len(obj.obj_what_changed()))
        self.assertEqual({'name': "My Name"}, obj.obj_get_changes())

    def test_obj_reset_changes_recursive(self):
        obj = TestObject()
        obj.id = "My ID"
        obj.name = "My Name"
        obj.nested = TestObject()
        obj.nested.id = "My ID"

        self.assertEqual(3, len(obj.obj_what_changed()))

        obj.obj_reset_changes()
        self.assertEqual(1, len(obj.obj_what_changed()))

        obj.obj_reset_changes(recursive=True)
        self.assertEqual(0, len(obj.obj_what_changed()))

    def test_obj_get_original_value(self):
        # Create an object
        obj = TestObject()
        obj.id = "My ID"
        obj.name = "My Name"

        # Rset one of the changes
        obj.obj_reset_changes(['id'])

        # Update the reset field
        obj.id = "My New ID"

        # Ensure the "current" value is correct
        self.assertEqual("My New ID", obj.id)

        # Ensure the "original" value is correct
        self.assertEqual("My ID", obj.obj_get_original_value('id'))
        self.assertEqual("My Name", obj.obj_get_original_value('name'))

        # Update the reset field again
        obj.id = "My New New ID"

        # Ensure the "current" value is correct
        self.assertEqual("My New New ID", obj.id)

        # Ensure the "original" value is still correct
        self.assertEqual("My ID", obj.obj_get_original_value('id'))
        self.assertEqual("My Name", obj.obj_get_original_value('name'))

        # Ensure a KeyError is raised when value exists
        with testtools.ExpectedException(KeyError):
            obj.obj_get_original_value('nested')

    def test_deepcopy(self):
        # Create the Original object
        o_obj = TestObject()
        o_obj.id = "My ID"
        o_obj.name = "My Name"

        # Clear the "changed" flag for one of the two fields we set
        o_obj.obj_reset_changes(['name'])

        # Deepcopy the object
        c_obj = copy.deepcopy(o_obj)

        # Ensure the copy was successful
        self.assertEqual(o_obj.id, c_obj.id)
        self.assertEqual(o_obj.name, c_obj.name)
        self.assertEqual(o_obj.obj_attr_is_set('nested'),
                         c_obj.obj_attr_is_set('nested'))

        self.assertEqual(o_obj.obj_get_changes(), c_obj.obj_get_changes())
        self.assertEqual(o_obj.to_primitive(), c_obj.to_primitive())

    def test_eq(self):
        # Create two equal objects
        obj_one = TestObject(id="My ID", name="My Name")
        obj_two = TestObject(id="My ID", name="My Name")

        # Ensure they evaluate to equal
        self.assertEqual(obj_one, obj_two)

        # Change a value on one object
        obj_two.name = 'Other Name'

        # Ensure they do not evaluate to equal
        self.assertNotEqual(obj_one, obj_two)

    def test_eq_false(self):
        obj = TestObject(id="My ID", name="My Name")
        self.assertFalse(obj == tuple())
        self.assertNotEqual(obj, tuple())

    def test_ne(self):
        # Create two equal objects
        obj_one = TestObject(id="My ID", name="My Name")
        obj_two = TestObject(id="My ID", name="My Name")

        # Ensure they evaluate to equal
        self.assertEqual(obj_one, obj_two)

        # Change a value on one object
        obj_two.name = 'Other Name'

        # Ensure they do not evaluate to equal
        self.assertNotEqual(obj_one, obj_two)


class DictObjectMixinTest(oslotest.base.BaseTestCase):
    def test_cast_to_dict(self):
        # Create an object
        obj = TestObjectDict()
        obj.id = 1
        obj.name = "My Name"

        expected = {
            'id': 1,
            'name': 'My Name',
        }

        self.assertEqual(expected, dict(obj))

    def test_gititem(self):
        obj = TestObjectDict(name=1)
        self.assertEqual(1, obj['name'])

    def test_setitem(self):
        obj = TestObjectDict()
        obj['name'] = 1
        self.assertEqual(1, obj.name)

    def test_contains(self):
        obj = TestObjectDict(name=1)
        self.assertIn('name', obj)

    def test_get(self):
        obj = TestObjectDict(name=1)
        v = obj.get('name')
        self.assertEqual(1, v)

    def test_get_missing(self):
        obj = TestObjectDict(name=1)
        self.assertFalse(obj.obj_attr_is_set('foo'))
        with testtools.ExpectedException(AttributeError):
            obj.get('foo')

    def test_get_default(self):
        obj = TestObjectDict(name='n')
        v = obj.get('name', value='default')
        self.assertEqual('n', v)

    def test_get_default_with_patch(self):
        obj = TestObjectDict(name='v')
        fname = 'designate.objects.base.DesignateObject.obj_attr_is_set'
        with mock.patch(fname) as attr_is_set:
            attr_is_set.return_value = False
            v = obj.get('name', value='default')
            self.assertEqual('default', v)

    def test_iteritems(self):
        obj = TestObjectDict(name=None, id=1)
        items = tuple(obj.items())
        self.assertEqual(
            [('id', 1), ('name', None)],
            sorted(items)
        )

    def test_jsonutils_to_primitive(self):
        obj = TestObjectDict(name="foo")
        dumped = jsonutils.to_primitive(obj, convert_instances=True)
        self.assertIsInstance(dumped, dict)
        self.assertEqual('foo', dumped['name'])


class ListObjectMixinTest(oslotest.base.BaseTestCase):
    def test_from_primitive(self):
        primitive = {
            'designate_object.name': 'TestObjectList',
            'designate_object.data': {
                'objects': [
                    {'designate_object.changes': ['id'],
                     'designate_object.data': {'id': 'One'},
                     'designate_object.name': 'TestObject',
                     'designate_object.namespace': 'designate',
                     'designate_object.version': '1.0'},
                    {'designate_object.changes': ['id'],
                     'designate_object.data': {'id': 'Two'},
                     'designate_object.name': 'TestObject',
                     'designate_object.namespace': 'designate',
                     'designate_object.version': '1.0'},
                ],
            },
            'designate_object.changes': ['objects'],
            'designate_object.namespace': 'designate',
            'designate_object.version': '1.0',
        }

        obj = objects.DesignateObject.from_primitive(primitive)

        self.assertEqual(2, len(obj))
        self.assertEqual(2, len(obj.objects))

        self.assertIsInstance(obj[0], TestObject)
        self.assertIsInstance(obj[1], TestObject)

        self.assertEqual('One', obj[0].id)
        self.assertEqual('Two', obj[1].id)

    def test_from_primitive_with_changes(self):
        primitive = {
            'designate_object.name': 'TestObjectList',
            'designate_object.data': {
                'objects': [
                    {
                        'designate_object.changes': ['id'],
                        'designate_object.data': {'id': 'One'},
                        'designate_object.name': 'TestObject',
                        'designate_object.namespace': 'designate',
                        'designate_object.version': '1.0'
                    },
                    {
                        'designate_object.changes': ['id'],
                        'designate_object.data': {'id': 'Two'},
                        'designate_object.name': 'TestObject',
                        'designate_object.namespace': 'designate',
                        'designate_object.version': '1.0'
                    },
                ],
            },
            'designate_object.changes': ['objects'],
            'designate_object.namespace': 'designate',
            'designate_object.version': '1.0',
        }

        obj = objects.DesignateObject.from_primitive(primitive)

        self.assertEqual(2, len(obj))
        self.assertEqual(2, len(obj.objects))

        self.assertIsInstance(obj[0], TestObject)
        self.assertIsInstance(obj[1], TestObject)

        self.assertEqual('One', obj[0].id)
        self.assertEqual('Two', obj[1].id)

        self.assertEqual(1, len(obj.obj_what_changed()))

    def test_from_primitive_no_changes(self):
        primitive = {
            'designate_object.name': 'TestObjectList',
            'designate_object.data': {
                'objects': [
                    {'designate_object.changes': [],
                     'designate_object.data': {'id': 'One'},
                     'designate_object.name': 'TestObject',
                     'designate_object.namespace': 'designate',
                     'designate_object.version': '1.0'},
                    {'designate_object.changes': [],
                     'designate_object.data': {'id': 'Two'},
                     'designate_object.name': 'TestObject',
                     'designate_object.namespace': 'designate',
                     'designate_object.version': '1.0'},
                ],
            },
            'designate_object.changes': [],
            'designate_object.namespace': 'designate',
            'designate_object.version': '1.0',
        }

        obj = objects.DesignateObject.from_primitive(primitive)

        self.assertEqual(2, len(obj))
        self.assertEqual(2, len(obj.objects))

        self.assertIsInstance(obj[0], TestObject)
        self.assertIsInstance(obj[1], TestObject)

        self.assertEqual('One', obj[0].id)
        self.assertEqual('Two', obj[1].id)

        self.assertEqual(0, len(obj.obj_what_changed()))

    def test_cast_to_list(self):
        # Create a few objects
        obj_one = TestObject()
        obj_one.id = "One"
        obj_two = TestObject()
        obj_two.id = "Two"

        # Create a ListObject
        obj = TestObjectList(objects=[obj_one, obj_two])

        expected = [obj_one, obj_two]
        self.assertEqual(expected, list(obj))

    def test_to_primitive(self):
        # Create a few objects
        obj_one = TestObject()
        obj_one.id = "One"
        obj_two = TestObject()
        obj_two.id = "Two"

        # Create a ListObject
        obj = TestObjectList(objects=[obj_one, obj_two])

        primitive = obj.to_primitive()
        expected = {
            'designate_object.name': 'TestObjectList',
            'designate_object.data': {
                'objects': [
                    {
                        'designate_object.changes': ['id'],
                        'designate_object.data': {'id': 'One'},
                        'designate_object.name': 'TestObject',
                        'designate_object.namespace': 'designate',
                        'designate_object.version': '1.0'
                    },
                    {
                        'designate_object.changes': ['id'],
                        'designate_object.data': {'id': 'Two'},
                        'designate_object.name': 'TestObject',
                        'designate_object.namespace': 'designate',
                        'designate_object.version': '1.0'
                    },
                ],
            },
            'designate_object.changes': ['objects'],
            'designate_object.namespace': 'designate',
            'designate_object.version': '1.0'
        }
        self.assertEqual(expected, primitive)

    def test_to_primitive_nested_obj(self):
        # Create a few objects
        obj_one = TestObject()
        obj_two = TestObject()
        obj_two.id = "Two"
        obj_one.nested = obj_two

        # Create a ListObject
        obj = TestObjectList(objects=[obj_one, obj_two])

        primitive = obj.to_primitive()
        expected = {
            'designate_object.name': 'TestObjectList',
            'designate_object.changes': ['objects'],
            'designate_object.data': {
                'objects': [
                    {
                        'designate_object.changes': ['nested'],
                        'designate_object.data': {'nested':
                            {
                                'designate_object.changes': ['id'],
                                'designate_object.data': {'id': 'Two'},
                                'designate_object.name': 'TestObject',
                                'designate_object.namespace': 'designate',
                                'designate_object.version': '1.0'}
                        },
                        'designate_object.name': 'TestObject',
                        'designate_object.namespace': 'designate',
                        'designate_object.version': '1.0'},
                    {
                        'designate_object.changes': ['id'],
                        'designate_object.data': {'id': 'Two'},
                        'designate_object.name': 'TestObject',
                        'designate_object.namespace': 'designate',
                        'designate_object.version': '1.0'
                    }
                ]
            },
            'designate_object.namespace': 'designate',
            'designate_object.version': '1.0'}

        self.assertEqual(expected, primitive)

    def test_obj_what_changed(self):
        # Create a few objects
        obj_one = TestObject()
        obj_two = TestObject()

        # Create a ListObject
        obj = TestObjectList(objects=[obj_one, obj_two])

        # Make sure there are no changes
        obj.obj_reset_changes()

        changes = obj.obj_what_changed()
        expected = set()

        self.assertEqual(expected, changes)

        # Make some changes
        obj_one.id = "One"
        obj_two.id = "Two"

        changes = obj.obj_what_changed()
        expected = {'objects'}

        self.assertEqual(expected, changes)

    def test_get_slice(self):
        # Create a few objects
        obj_one = TestObject(id="One")
        obj_two = TestObject(id="Two")

        # Create a ListObject
        obj = TestObjectList(objects=[obj_one, obj_two])

        theslice = obj[1:]
        expected = TestObjectList(objects=[obj_two])

        self.assertEqual(expected.objects, theslice.objects)
        self.assertNotEqual(obj.objects, theslice.objects)

    def test_setitem(self):
        # Create a few objects
        obj_one = TestObject(id="One")
        obj_two = TestObject(id="Two")

        # Create a ListObject
        obj = TestObjectList(objects=[obj_one, obj_two])

        obj[1] = obj_one

        self.assertEqual(obj.objects, [obj_one, obj_one])

    def test_contains(self):
        # Create a few objects
        obj_one = TestObject(id="One")
        obj_two = TestObject(id="Two")
        obj_three = TestObject(id="Three")

        # Create a ListObject
        obj = TestObjectList(objects=[obj_one, obj_two])

        self.assertIn(obj_one, obj)
        self.assertIn(obj_two, obj)
        self.assertNotIn(obj_three, obj)

    def test_extend(self):
        # Create a few objects
        obj_one = TestObject(id="One")
        obj_two = TestObject(id="Two")
        obj_three = TestObject(id="Three")

        # Create a ListObject
        ext_obj = TestObjectList(objects=[obj_one])
        obj = TestObjectList(objects=[obj_one, obj_two, obj_three])

        ext_obj.extend([obj_two, obj_three])

        self.assertEqual(obj.objects, ext_obj.objects)

    def test_insert(self):
        # Create a few objects
        obj_one = TestObject(id="One")
        obj_two = TestObject(id="Two")
        obj_three = TestObject(id="Three")

        # Create a ListObject
        obj = TestObjectList(objects=[obj_one, obj_three])

        obj.insert(1, obj_two)

        self.assertEqual([obj_one, obj_two, obj_three], obj.objects)

    def test_remove(self):
        # Create a few objects
        obj_one = TestObject(id="One")
        obj_two = TestObject(id="Two")

        # Create a ListObject
        obj = TestObjectList(objects=[obj_one, obj_two])
        obj.remove(obj_one)
        self.assertEqual([obj_two], obj.objects)

    def test_index(self):
        # Create a few objects
        obj_one = TestObject(id="One")
        obj_two = TestObject(id="Two")
        obj_three = TestObject(id="Three")

        # Create a ListObject
        obj = TestObjectList(objects=[obj_one, obj_two, obj_three])

        self.assertEqual(1, obj.index(obj_two))

    def test_count(self):
        # Create a few objects
        obj_one = TestObject(id="One")
        obj_two = TestObject(id="Two")

        # Create a ListObject
        obj = TestObjectList(objects=[obj_one, obj_two, obj_two])

        self.assertEqual(2, obj.count(obj_two))

    def test_sort(self):
        # Create a few objects
        obj_one = TestObject(id=1)
        obj_two = TestObject(id=2)
        obj_three = TestObject(id=3)

        # Create a ListObject
        obj = TestObjectList(objects=[obj_two, obj_three, obj_one])
        obj.sort(key=attrgetter('id'))

        self.assertEqual([obj_one, obj_two, obj_three], obj.objects)

    def test_to_dict_list_mixin(self):
        # Create a ListObject containing an ObjectList
        obj = TestObjectList(objects=[TestObject()])

        dict_ = obj.to_dict()
        expected = {'objects': [{}]}
        self.assertEqual(expected, dict_)

    def test_to_list(self):
        # Create a few objects
        obj_one = TestObject(id="One")
        obj_three = TestObject(id="Three")

        # Create a ListObject
        obj = TestObjectList(objects=[obj_one, obj_three])

        li = obj.to_list()
        self.assertEqual([{'id': 'One'}, {'id': 'Three'}], li)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""shared_zones

Revision ID: b20189fd288e
Revises: e5e2199ed76e
Create Date: 2022-09-22 20:50:03.056609

"""
from alembic import op
from oslo_utils import uuidutils
import sqlalchemy as sa

from designate.storage.sqlalchemy.types import UUID

# revision identifiers, used by Alembic.
revision = 'b20189fd288e'
down_revision = 'e5e2199ed76e'
branch_labels = None
depends_on = None


def upgrade() -> None:
    meta = sa.MetaData()

    op.create_table(
        'shared_zones', meta,
        sa.Column('id', UUID,
                  default=uuidutils.generate_uuid, primary_key=True),
        sa.Column('created_at', sa.DateTime),
        sa.Column('updated_at', sa.DateTime),
        sa.Column('zone_id', UUID, nullable=False),
        sa.Column('project_id', sa.String(36), nullable=False),
        sa.Column('target_project_id', sa.String(36), nullable=False),

        sa.UniqueConstraint('zone_id', 'project_id', 'target_project_id',
                            name='unique_shared_zone'),
        sa.ForeignKeyConstraint(['zone_id'], ['zones.id'], ondelete='CASCADE'),
        mysql_engine='InnoDB',
        mysql_charset='utf8',
    )



# Copyright 2016 Hewlett Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging

from designate import exceptions
from designate import objects
from designate import policy
from designate.scheduler.filters import base

LOG = logging.getLogger(__name__)


class PoolIDAttributeFilter(base.Filter):
    """This allows users with the correct role to specify the exact pool_id
    to schedule the supplied zone to.

    This is supplied as an attribute on the zone

    .. code-block:: python
        :emphasize-lines: 3

        {
            "attributes": {
                "pool_id": "794ccc2c-d751-44fe-b57f-8894c9f5c842"
            },
            "email": "user@example.com",
            "name": "example.com."
        }

    The pool is loaded to ensure it exists, and then a policy check is
    performed to ensure the user has the correct role.

    .. warning::

        This should only be enabled if required, as it will raise a
        403 Forbidden if a user without the correct role uses it.
    """

    name = 'pool_id_attribute'
    """Name to enable in the ``[designate:central:scheduler].filters`` option
    list
    """

    def filter(self, context, pools, zone):
        """Attempt to load and set the pool to the one provided in the
        Zone attributes.

        :param context: :class:`designate.context.DesignateContext` - Context
            Object from request
        :param pools: :class:`designate.objects.pool.PoolList` - List of pools
            to choose from
        :param zone: :class:`designate.objects.zone.Zone` - Zone to be created
        :return: :class:`designate.objects.pool.PoolList` -- A PoolList with
            containing a single pool.
        :raises: Forbidden, PoolNotFound
        """

        try:
            if zone.attributes.get('pool_id'):
                pool_id = zone.attributes.get('pool_id')
                try:
                    pool = self.storage.get_pool(context, pool_id)
                except Exception:
                    return objects.PoolList()
                policy.check('zone_create_forced_pool', context, pool)
                if pool in pools:
                    pools = objects.PoolList()
                    pools.append(pool)
                return pools
            else:
                return pools
        except exceptions.RelationNotLoaded:
            return pools



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import copy
import functools
import importlib.resources
import inspect
import os
import socket
import sys

from jinja2 import Template
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils.netutils import is_valid_ipv6
from oslo_utils import uuidutils

from designate.common import config
import designate.conf
from designate import exceptions
from designate.i18n import _

LOG = logging.getLogger(__name__)
CONF = designate.conf.CONF

# Default datetime format

DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'


def find_config(config_path):
    """
    Find a configuration file using the given hint.

    Code nabbed from cinder.

    :param config_path: Full or relative path to the config.
    :returns: List of config paths
    """
    possible_locations = [
        config_path,
        os.path.join(CONF.pybasedir, "etc", "designate", config_path),
        os.path.join(CONF.pybasedir, "etc", config_path),
        os.path.join(CONF.pybasedir, config_path),
        "/etc/designate/%s" % config_path,
    ]

    found_locations = []

    for path in possible_locations:
        LOG.debug('Searching for configuration at path: %s' % path)
        if os.path.exists(path):
            LOG.debug('Found configuration at path: %s' % path)
            found_locations.append(os.path.abspath(path))

    return found_locations


def read_config(prog, argv):
    logging.register_options(CONF)
    config_files = find_config('%s.conf' % prog)
    cfg.CONF(argv[1:], project='designate', prog=prog,
             default_config_files=config_files)
    config.set_defaults()


def resource_string(*args):
    if len(args) == 0:
        raise ValueError()

    resource_path = importlib.resources.files(
        'designate').joinpath('resources', *args)

    if not resource_path.exists():
        raise exceptions.ResourceNotFound('Could not find the requested '
                                          'resource: %s' % resource_path)

    return resource_path.read_bytes()


def load_schema(version, name):
    schema_string = resource_string('schemas', version, '%s.json' % name)

    return jsonutils.loads(schema_string.decode('utf-8'))


def load_template(template_name):
    template_string = resource_string('templates', template_name)

    return Template(template_string.decode('utf-8'),
                    keep_trailing_newline=True)


def render_template(template, **template_context):
    if not isinstance(template, Template):
        template = load_template(template)

    return template.render(**template_context)


def execute(*cmd, **kw):
    """Execute a command in a subprocess, blocking.
    """
    root_helper = kw.pop('root_helper', CONF.root_helper)
    run_as_root = kw.pop('run_as_root', True)
    return processutils.execute(*cmd, run_as_root=run_as_root,
                                root_helper=root_helper, **kw)


def deep_dict_merge(a, b):
    if not isinstance(b, dict):
        return b

    result = copy.deepcopy(a)

    for k, v in b.items():
        if k in result and isinstance(result[k], dict):
            result[k] = deep_dict_merge(result[k], v)
        else:
            result[k] = copy.deepcopy(v)

    return result


def validate_uuid(*check):
    """
    A wrapper to ensure that API controller methods arguments are valid UUID's.

    Usage:
    @validate_uuid('zone_id')
    def get_all(self, zone_id):
        return {}
    """
    def inner(f):
        def wrapper(*args, **kwargs):
            arg_spec = inspect.getfullargspec(f).args

            # Ensure that we have the exact number of parameters that the
            # function expects.  This handles URLs like
            # /v2/zones/<UUID - valid or invalid>/invalid
            # get, patch and delete return a 404, but Pecan returns a 405
            # for a POST at the same URL
            if len(arg_spec) != len(args):
                raise exceptions.NotFound()

            # Ensure that we have non-empty parameters in the cases where we
            # have sub controllers - i.e. controllers at the 2nd level
            # This is for URLs like /v2/zones/nameservers
            # Ideally Pecan should be handling these cases, but until then
            # we handle those cases here.
            if len(args) <= len(check):
                raise exceptions.NotFound()

            for name in check:
                pos = arg_spec.index(name)
                if not uuidutils.is_uuid_like(args[pos]):
                    raise exceptions.InvalidUUID(
                        f'Invalid UUID {name}: {args[pos]}'
                    )
            return f(*args, **kwargs)
        return functools.wraps(f)(wrapper)
    return inner


def get_proxies():
    """Return a requests compatible dict like seen here
    http://docs.python-requests.org/en/latest/user/advanced/#proxies for
    consumption in clients when we need to proxy requests.
    """
    proxies = {}
    if CONF.proxy.no_proxy:
        proxies['no_proxy'] = CONF.proxy.no_proxy
    if CONF.proxy.http_proxy is not None:
        proxies['http'] = CONF.proxy.http_proxy

    if CONF.proxy.https_proxy is not None:
        proxies['https'] = CONF.proxy.https_proxy
    elif 'http' in proxies:
        proxies['https'] = proxies['http']

    return proxies


def cache_result(function):
    """A function decorator to cache the result of the first call, every
    additional call will simply return the cached value.

    If we were python3 only, we would have used functools.lru_cache() in place
    of this. If there's a python2 backport in a lightweight library, then we
    should switch to that.
    """
    # NOTE: We're cheating a little here, by using a mutable type (a list),
    #       we're able to read and update the value from within in inline
    #       wrapper method. If we used an immutable type, the assignment
    #       would not work as we want.
    cache = []

    def wrapper(cls_instance):
        if not cache:
            cache.append(function(cls_instance))
        return cache[0]
    return wrapper


def split_host_port(string, default_port=53):
    try:
        (host, port) = string.split(':', 1)
        port = int(port)
    except ValueError:
        host = str(string)
        port = default_port

    return (host, port)


def get_paging_params(context, params, sort_keys):
    """
    Extract any paging parameters
    """
    marker = params.pop('marker', None)
    limit = params.pop('limit', CONF['service:api'].default_limit_v2)
    sort_key = params.pop('sort_key', None)
    sort_dir = params.pop('sort_dir', None)
    max_limit = CONF['service:api'].max_limit_v2

    if isinstance(limit, str) and limit.lower() == "max":
        # Support for retrieving the max results at once. If set to "max",
        # the configured max limit will be used.
        limit = max_limit

    elif limit:
        # Negative and zero limits are not caught in storage.
        # With a number bigger than MAXSIZE, rpc throws an 'OverflowError long
        # too big to convert'.
        # So the parameter 'limit' is checked here.
        invalid_limit_message = ('limit should be an integer between 1 and '
                                 '%(max)s' % {'max': max_limit})
        try:
            int_limit = int(limit)
            if int_limit <= 0 or int_limit > sys.maxsize:
                raise exceptions.InvalidLimit(invalid_limit_message)
        # This exception is raised for non ints when int(limit) is called
        except ValueError:
            raise exceptions.InvalidLimit(invalid_limit_message)

    # sort_dir is checked in paginate_query.
    # We duplicate the sort_dir check here to throw a more specific
    # exception than ValueError.
    if sort_dir and sort_dir not in ['asc', 'desc']:
        raise exceptions.InvalidSortDir(_("Unknown sort direction, "
                                          "must be 'desc' or 'asc'"))

    if sort_keys is None:
        sort_key = None
        sort_dir = None

    elif sort_key and sort_key not in sort_keys:
        msg = f'sort key must be one of {sort_keys}'
        raise exceptions.InvalidSortKey(msg)
    elif sort_key == 'tenant_id' and not context.all_tenants:
        sort_key = None

    return marker, limit, sort_key, sort_dir


def bind_tcp(host, port, tcp_backlog, tcp_keepidle=None):
    """Bind to a TCP port and listen.
    Use reuseaddr, reuseport if available, keepalive if specified

    :param host: IPv4/v6 address or "". "" binds to every IPv4 interface.
    :type host: str
    :param port: TCP port
    :type port: int
    :param tcp_backlog: TCP listen backlog
    :type tcp_backlog: int
    :param tcp_keepidle: TCP keepalive interval
    :type tcp_keepidle: int
    :returns: socket
    """
    LOG.info('Opening TCP Listening Socket on %(host)s:%(port)d',
             {'host': host, 'port': port})
    family = socket.AF_INET6 if is_valid_ipv6(host) else socket.AF_INET
    sock_tcp = socket.socket(family, socket.SOCK_STREAM)
    sock_tcp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock_tcp.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

    # NOTE: Linux supports socket.SO_REUSEPORT only in 3.9 and later releases.
    try:
        sock_tcp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    except Exception:
        LOG.info('SO_REUSEPORT not available, ignoring.')

    # This option isn't available in the OS X version of eventlet
    if tcp_keepidle and hasattr(socket, 'TCP_KEEPIDLE'):
        sock_tcp.setsockopt(socket.IPPROTO_TCP,
                            socket.TCP_KEEPIDLE,
                            tcp_keepidle)

    sock_tcp.settimeout(1)
    sock_tcp.bind((host, port))
    if port == 0:
        newport = sock_tcp.getsockname()[1]
        LOG.info('Listening on TCP port %(port)d', {'port': newport})

    sock_tcp.listen(tcp_backlog)

    return sock_tcp


def bind_udp(host, port):
    """Bind to an UDP port and listen.
    Use reuseaddr, reuseport if available

    :param host: IPv4/v6 address or "". "" binds to every IPv4 interface.
    :type host: str
    :param port: UDP port
    :type port: int
    :returns: socket
    """
    LOG.info('Opening UDP Listening Socket on %(host)s:%(port)d',
             {'host': host, 'port': port})
    family = socket.AF_INET6 if is_valid_ipv6(host) else socket.AF_INET
    sock_udp = socket.socket(family, socket.SOCK_DGRAM)
    sock_udp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    # NOTE: Linux supports socket.SO_REUSEPORT only in 3.9 and later releases.
    try:
        sock_udp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    except Exception:
        LOG.info('SO_REUSEPORT not available, ignoring.')

    sock_udp.settimeout(1)
    sock_udp.bind((host, port))
    if port == 0:
        newport = sock_udp.getsockname()[1]
        LOG.info('Listening on UDP port %(port)d', {'port': newport})

    return sock_udp


def max_prop_time(timeout, max_retries, retry_interval, delay):
    return timeout * max_retries + max_retries * retry_interval + delay



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Author: Federico Ceratto <federico.ceratto@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

from oslo_config import fixture as cfg_fixture
import oslotest.base

import designate.conf
from designate import exceptions
from designate import policy
from designate.producer import service
from designate import rpc
import designate.service
from designate.tests import base_fixtures


CONF = designate.conf.CONF


@mock.patch.object(service.rpcapi.CentralAPI, 'get_instance', mock.Mock())
class ProducerServiceTest(oslotest.base.BaseTestCase):
    @mock.patch.object(policy, 'init')
    @mock.patch.object(rpc, 'get_client')
    @mock.patch.object(rpc, 'init')
    @mock.patch.object(rpc, 'initialized')
    def setUp(self, mock_rpc_initialized, mock_rpc_init, mock_rpc_get_client,
              mock_policy_init):
        super().setUp()

        self.useFixture(cfg_fixture.Config(CONF))
        self.stdlog = base_fixtures.StandardLogging()
        self.useFixture(self.stdlog)

        mock_rpc_initialized.return_value = False

        self.tg = mock.Mock()
        self.service = service.Service()
        self.service.coordination = mock.Mock()
        self.service.tg = self.tg

        mock_policy_init.assert_called_once()
        mock_rpc_initialized.assert_called_once()
        mock_rpc_init.assert_called_once()

    @mock.patch.object(rpc, 'get_notifier', mock.Mock())
    @mock.patch.object(service.coordination, 'Partitioner')
    @mock.patch.object(designate.service.RPCService, 'start')
    def test_service_start(self, mock_rpc_start, mock_partitioner):
        CONF.set_override('enabled_tasks', None, 'service:producer')

        mock_partition = mock.Mock()
        mock_partitioner.return_value = mock_partition

        self.service.start()

        mock_rpc_start.assert_called_with()
        mock_partition.watch_partition_change.assert_called()
        mock_partition.start.assert_called()

        # Make sure that tasks were added to the tg timer.
        self.tg.add_timer_args.assert_called()
        self.assertEqual(6, self.tg.add_timer_args.call_count)

    @mock.patch.object(service.coordination, 'Partitioner')
    @mock.patch.object(designate.service.RPCService, 'start')
    def test_service_start_all_extension_disabled(self, mock_rpc_start,
                                                  mock_partitioner):
        CONF.set_override('enabled_tasks', [], 'service:producer')
        self.assertRaisesRegex(
            exceptions.ConfigurationError,
            r'No periodic tasks found matching: \[\]',
            self.service.start,
        )

        CONF.set_override('enabled_tasks', ['None'], 'service:producer')
        self.assertRaisesRegex(
            exceptions.ConfigurationError,
            r'No periodic tasks found matching: \[\'None\'\]',
            self.service.start,
        )

    def test_service_stop(self):
        self.service.stop()

        self.service.coordination.stop.assert_called()

        self.assertIn('Stopping producer service', self.stdlog.logger.output)

    def test_service_name(self):
        self.assertEqual('producer', self.service.service_name)

    @mock.patch.object(policy, 'init', mock.Mock())
    @mock.patch.object(rpc, 'get_client', mock.Mock())
    @mock.patch.object(rpc, 'init', mock.Mock())
    @mock.patch.object(rpc, 'initialized', mock.Mock(return_value=False))
    @mock.patch.object(rpc, 'get_notifier', mock.Mock())
    def test_producer_rpc_topic(self):
        CONF.set_override('topic', 'test-topic', 'service:producer')

        self.service = service.Service()

        self.assertEqual('test-topic', self.service.rpc_topic)
        self.assertEqual('producer', self.service.service_name)

    def test_central_api(self):
        self.assertIsInstance(self.service.central_api, mock.Mock)



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_log import log as logging

from designate import objects
from designate.objects.adapters.yaml import base

LOG = logging.getLogger(__name__)


class PoolYAMLAdapter(base.YAMLAdapter):
    ADAPTER_OBJECT = objects.Pool
    MODIFICATIONS = {
        'fields': {
            'id': {
                'read_only': False
            },
            'name': {
                'read_only': False
            },
            'description': {
                'read_only': False
            },
            'attributes': {
                'read_only': False
            },
            'ns_records': {
                'read_only': False
            },
            'nameservers': {
                'read_only': False
            },
            'targets': {
                'read_only': False
            },
            'also_notifies': {
                'read_only': False
            },
            'catalog_zone': {
                'read_only': False
            }
        }
    }


class PoolListYAMLAdapter(base.YAMLAdapter):
    ADAPTER_OBJECT = objects.PoolList
    MODIFICATIONS = {}



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Graham Hayes <graham.hayes@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from designate.objects import base
from designate.objects import fields


@base.DesignateRegistry.register
class ZoneTransferAccept(base.DictObjectMixin, base.PersistentObjectMixin,
                         base.DesignateObject):
    fields = {
        'zone_transfer_request_id': fields.UUIDFields(nullable=True),
        'tenant_id': fields.StringFields(nullable=True),
        'status': fields.EnumField(nullable=True, valid_values=[
            'ACTIVE', 'PENDING', 'DELETED', 'ERROR', 'COMPLETE'
        ]),
        'key': fields.StringFields(maxLength=160),
        'zone_id': fields.UUIDFields(nullable=True),
    }

    STRING_KEYS = [
        'id', 'zone_id', 'tenant_id', 'zone_transfer_request_id'
    ]


@base.DesignateRegistry.register
class ZoneTransferAcceptList(base.ListObjectMixin, base.DesignateObject,
                             base.PagedListObjectMixin):
    LIST_ITEM_TYPE = ZoneTransferAccept

    fields = {
        'objects': fields.ListOfObjectsField('ZoneTransferAccept'),
    }



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import functools
import threading
import time

from oslo_db import exception as db_exception
from oslo_log import log as logging
from oslo_utils import excutils

from designate.storage import sql
from designate.storage import sqlalchemy


LOG = logging.getLogger(__name__)
RETRY_STATE = threading.local()


def get_storage():
    """Return the engine class"""
    return sqlalchemy.SQLAlchemyStorage()


def _retry_on_deadlock(exc):
    """Filter to trigger retry a when a Deadlock is received."""
    # TODO(kiall): This is a total leak of the SQLA Driver, we'll need a better
    #              way to handle this.
    if isinstance(exc, db_exception.DBDeadlock):
        LOG.warning("Deadlock detected. Retrying...")
        return True
    return False


def retry(cb=None, retries=150, delay=50, deep_copy=True):
    """A retry decorator that ignores attempts at creating nested retries"""
    def outer(f):
        @functools.wraps(f)
        def retry_wrapper(self, *args, **kwargs):
            if not hasattr(RETRY_STATE, 'held'):
                # Create the state vars if necessary
                RETRY_STATE.held = False
                RETRY_STATE.retries = 0

            if not RETRY_STATE.held:
                # We're the outermost retry decorator
                RETRY_STATE.held = True

                try:
                    while True:
                        try:
                            if deep_copy:
                                result = f(self, *copy.deepcopy(args),
                                           **copy.deepcopy(kwargs))
                            else:
                                # perform shallow copy
                                result = f(self, *copy.copy(args),
                                           **copy.copy(kwargs))
                            break
                        except Exception as exc:
                            RETRY_STATE.retries += 1
                            if RETRY_STATE.retries >= retries:
                                # Exceeded retry attempts, raise.
                                raise
                            elif cb is not None and cb(exc) is False:
                                # We're not setup to retry on this exception.
                                raise
                            else:
                                # Retry, with a delay.
                                time.sleep(delay / float(1000))

                finally:
                    RETRY_STATE.held = False
                    RETRY_STATE.retries = 0

            else:
                # We're an inner retry decorator, just pass on through.
                if deep_copy:
                    result = f(self, *copy.deepcopy(args),
                               **copy.deepcopy(kwargs))
                else:
                    # perform shallow copy
                    result = f(self, *copy.copy(args),
                               **copy.copy(kwargs))

            return result
        retry_wrapper.__wrapped_function = f
        retry_wrapper.__wrapper_name = 'retry'
        return retry_wrapper
    return outer


def transaction(f):
    """Transaction decorator, to be used on class instances with a
    self.storage attribute
    """

    @retry(cb=_retry_on_deadlock)
    @functools.wraps(f)
    def transaction_wrapper(self, *args, **kwargs):
        with sql.get_write_session() as session:
            # session.begin()
            try:
                result = f(self, *args, **kwargs)
                # session.commit()
                return result
            except Exception:
                with excutils.save_and_reraise_exception():
                    session.rollback()

    transaction_wrapper.__wrapped_function = f
    transaction_wrapper.__wrapper_name = 'transaction'
    return transaction_wrapper


def transaction_shallow_copy(f):
    """Transaction decorator, to be used on class instances with a
    self.storage attribute where shallow copy of args, kwargs is used
    """
    @retry(cb=_retry_on_deadlock, deep_copy=False)
    @functools.wraps(f)
    def transaction_wrapper(self, *args, **kwargs):
        with sql.get_write_session() as session:
            # session.begin()
            try:
                result = f(self, *args, **kwargs)
                # session.commit()
                return result
            except Exception:
                with excutils.save_and_reraise_exception():
                    session.rollback()

    transaction_wrapper.__wrapped_function = f
    transaction_wrapper.__wrapper_name = 'transaction_shallow_copy'
    return transaction_wrapper



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_config import fixture as cfg_fixture
import oslotest.base
from unittest import mock

from designate.api import wsgi
import designate.conf


CONF = designate.conf.CONF


class TestApiWsgi(oslotest.base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(cfg_fixture.Config(CONF))

    @mock.patch('paste.deploy.loadapp')
    @mock.patch('designate.heartbeat_emitter.get_heartbeat_emitter')
    @mock.patch('designate.common.profiler.setup_profiler')
    @mock.patch('designate.rpc.init')
    @mock.patch('designate.rpc.initialized')
    @mock.patch('designate.policy.init')
    @mock.patch('oslo_log.log.setup')
    @mock.patch('oslo_config.cfg.CONF')
    def test_init_application(self, mock_cfg, mock_log_setup, mock_policy_init,
                              mock_rpc_initialized,
                              mock_rpc_init,
                              mock_setup_profiler,
                              mock_get_heartbeat_emitter,
                              mock_loadapp):
        CONF.set_override('host', 'foo')
        mock_rpc_initialized.return_value = False

        wsgi.init_application()

        mock_cfg.assert_called_with(
            [], project='designate',
            default_config_files=['/etc/designate/api-paste.ini',
                                  '/etc/designate/designate.conf']
        )
        mock_log_setup.assert_called_with(mock.ANY, 'designate')
        mock_policy_init.assert_called()
        mock_rpc_init.assert_called()
        mock_setup_profiler.assert_called_with('designate-api', 'foo')
        mock_get_heartbeat_emitter.assert_called_with('api')
        mock_loadapp.assert_called_with(
            'config:/etc/designate/api-paste.ini', name='osapi_dns'
        )

    def test_get_config_files(self):
        self.assertEqual(
            ['/etc/designate/api-paste.ini', '/etc/designate/designate.conf'],
            wsgi._get_config_files()
        )

    def test_get_config_files_with_custom_env(self):
        env = {'OS_DESIGNATE_CONFIG_DIR': '/opt/designate'}

        self.assertEqual(
            ['/opt/designate/api-paste.ini', '/opt/designate/designate.conf'],
            wsgi._get_config_files(env)
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add catalog zones

Revision ID: 9099de8ae11c
Revises: a005af3aa38e
Create Date: 2023-05-15 09:30:11.476307

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '9099de8ae11c'
down_revision = 'a005af3aa38e'
branch_labels = None
depends_on = None


def upgrade() -> None:
    ZONE_TYPES = ['PRIMARY', 'SECONDARY', 'CATALOG']

    with op.batch_alter_table('zones') as batch_op:
        batch_op.alter_column('type', type_=sa.Enum(name='type',
                                                    *ZONE_TYPES),
                              existing_type=sa.Enum, existing_nullable=False)



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import copy

from keystoneauth1.access import service_catalog as ksa_service_catalog
from keystoneauth1 import plugin
from oslo_context import context
from oslo_log import log as logging

from designate import policy

LOG = logging.getLogger(__name__)


class DesignateContext(context.RequestContext):

    _all_tenants = False
    _hide_counts = False
    _abandon = None
    original_project_id = None
    _edit_managed_records = False
    _hard_delete = False
    _client_addr = None
    _delete_shares = False
    FROM_DICT_EXTRA_KEYS = [
        'original_project_id', 'service_catalog', 'all_tenants', 'abandon',
        'edit_managed_records', 'tsigkey_id', 'hide_counts', 'client_addr',
        'hard_delete', 'delete_shares'
    ]

    def __init__(self, service_catalog=None, all_tenants=False, abandon=None,
                 tsigkey_id=None, original_project_id=None,
                 edit_managed_records=False, hide_counts=False,
                 client_addr=None, user_auth_plugin=None,
                 hard_delete=False, delete_shares=False, **kwargs):
        super().__init__(**kwargs)

        self.user_auth_plugin = user_auth_plugin
        self.service_catalog = service_catalog
        self.tsigkey_id = tsigkey_id

        self.original_project_id = original_project_id

        self.all_tenants = all_tenants
        self.abandon = abandon
        self.edit_managed_records = edit_managed_records
        self.hard_delete = hard_delete
        self.hide_counts = hide_counts
        self.client_addr = client_addr
        self.delete_shares = delete_shares

    def deepcopy(self):
        return self.from_dict(self.to_dict())

    def to_dict(self):
        d = super().to_dict()

        # Override the user_identity field to account for TSIG. When a TSIG key
        # is used as authentication e.g. via MiniDNS, it will act as a form
        # of "user",
        user = self.user_id or '-'

        if self.tsigkey_id and not self.user_id:
            user = 'TSIG:%s' % self.tsigkey_id

        user_idt = (
            self.user_idt_format.format(
                user=user,
                project_id=self.project_id or '-',
                domain=self.domain_id or '-',
                user_domain=self.user_domain_id or '-',
                p_domain=self.project_domain_id or '-')
        )

        # Update the dict with Designate specific extensions and overrides
        d.update({
            'user_identity': user_idt,
            'original_project_id': self.original_project_id,
            'service_catalog': self.service_catalog,
            'all_tenants': self.all_tenants,
            'abandon': self.abandon,
            'edit_managed_records': self.edit_managed_records,
            'hard_delete': self.hard_delete,
            'tsigkey_id': self.tsigkey_id,
            'hide_counts': self.hide_counts,
            'client_addr': self.client_addr,
            'delete_shares': self.delete_shares,
        })

        return copy.deepcopy(d)

    def elevated(self, show_deleted=None, all_tenants=False,
                 edit_managed_records=False, hard_delete=False):
        """Return a version of this context with admin flag set.
        Optionally set all_tenants and edit_managed_records
        """
        context = self.deepcopy()
        context.is_admin = True

        # NOTE(kiall): Ugly - required to match http://tinyurl.com/o3y8qmw
        context.roles.append('admin')

        if show_deleted is not None:
            context.show_deleted = show_deleted

        if all_tenants:
            context.all_tenants = True

        if edit_managed_records:
            context.edit_managed_records = True

        if hard_delete:
            context.hard_delete = True

        return context

    def sudo(self, project_id):

        policy.check('use_sudo', self)

        LOG.info('Accepted sudo from user %(user)s to project_id %(project)s',
                 {'user': self.user_id, 'project': project_id})
        self.original_project_id = self.project_id
        self.project_id = project_id

    @classmethod
    def get_admin_context(cls, **kwargs):
        # TODO(kiall): Remove Me
        kwargs['is_admin'] = True
        kwargs['roles'] = ['admin', 'reader']

        return cls(None, **kwargs)

    @property
    def all_tenants(self):
        return self._all_tenants

    @all_tenants.setter
    def all_tenants(self, value):
        if value:
            policy.check('all_tenants', self)
        self._all_tenants = value

    @property
    def hide_counts(self):
        return self._hide_counts

    @hide_counts.setter
    def hide_counts(self, value):
        self._hide_counts = value

    @property
    def abandon(self):
        return self._abandon

    @abandon.setter
    def abandon(self, value):
        if value:
            policy.check('abandon_zone', self)
        self._abandon = value

    @property
    def edit_managed_records(self):
        return self._edit_managed_records

    @edit_managed_records.setter
    def edit_managed_records(self, value):
        if value:
            policy.check('edit_managed_records', self)
        self._edit_managed_records = value

    @property
    def hard_delete(self):
        return self._hard_delete

    @hard_delete.setter
    def hard_delete(self, value):
        if value:
            policy.check('hard_delete', self)
        self._hard_delete = value

    @property
    def client_addr(self):
        return self._client_addr

    @client_addr.setter
    def client_addr(self, value):
        self._client_addr = value

    @property
    def delete_shares(self):
        return self._delete_shares

    @delete_shares.setter
    def delete_shares(self, value):
        self._delete_shares = value

    def get_auth_plugin(self):
        if self.user_auth_plugin:
            return self.user_auth_plugin
        return _ContextAuthPlugin(self.auth_token, self.service_catalog)


class _ContextAuthPlugin(plugin.BaseAuthPlugin):
    """A keystoneauth auth plugin that uses the values from the Context.
    Ideally we would use the plugin provided by auth_token middleware however
    this plugin isn't serialized yet, so we construct one from the serialized
    auth data.
    """
    def __init__(self, auth_token, sc):
        super().__init__()

        self.auth_token = auth_token
        self.service_catalog = ksa_service_catalog.ServiceCatalogV2(sc)

    def get_token(self, *args, **kwargs):
        return self.auth_token

    def get_endpoint(self, session, service_type=None, interface=None,
                     region_name=None, service_name=None, **kwargs):
        endpoint = self.service_catalog.url_for(
            service_type=service_type, service_name=service_name,
            interface=interface, region_name=region_name
        )
        return self.get_endpoint_data(session, endpoint_override=endpoint).url


def get_current():
    return context.get_current()



# Copyright 2015 Red Hat, Inc.
# Copyright 2017 Fujitsu Vietnam Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import dns.exception
from dns import ipv4
import re
import uuid

from oslo_versionedobjects import fields as ovoo_fields

from designate.common import constants


class IntegerField(ovoo_fields.IntegerField):
    pass


class BooleanField(ovoo_fields.BooleanField):
    pass


class PolymorphicObject(ovoo_fields.Object):
    def coerce(self, obj, attr, value):
        return value


class PolymorphicObjectField(ovoo_fields.AutoTypedField):
    def __init__(self, objtype, subclasses=False, **kwargs):
        self.AUTO_TYPE = PolymorphicObject(objtype, subclasses)
        self.objname = objtype
        super().__init__(**kwargs)


class ListOfObjectsField(ovoo_fields.ListOfObjectsField):
    pass


class ObjectFields(ovoo_fields.ObjectField):
    def __init__(self, objtype, subclasses=False, relation=False, **kwargs):
        self.AUTO_TYPE = ovoo_fields.List(
            ovoo_fields.Object(objtype, subclasses))
        self.objname = objtype
        super().__init__(objtype, **kwargs)
        self.relation = relation


class IntegerFields(IntegerField):
    def __init__(self, nullable=False, default=ovoo_fields.UnspecifiedDefault,
                 read_only=False, minimum=0, maximum=None):
        super().__init__(
            nullable=nullable, default=default, read_only=read_only
        )
        self.min = minimum
        self.max = maximum

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        if value is None:
            return value
        if value < self.min:
            # return self.min
            raise ValueError(f'Value must be >= {self.min} for field {attr}')
        if self.max and value > self.max:
            raise ValueError(f'Value too high for {attr}')
        return value


class StringFields(ovoo_fields.StringField):
    def __init__(self, nullable=False, read_only=False,
                 default=ovoo_fields.UnspecifiedDefault, description='',
                 maxLength=None):
        super().__init__(
            nullable=nullable, default=default, read_only=read_only
        )
        self.description = description
        self.maxLength = maxLength

    def coerce(self, obj, attr, value):
        if value is None:
            return self._null(obj, attr)
        else:
            value = super().coerce(obj, attr, value)
            if self.maxLength and len(value) > self.maxLength:
                raise ValueError(f'Value too long for {attr}')
            return value


class UUID(ovoo_fields.UUID):
    def coerce(self, obj, attr, value):
        try:
            value = int(value)
            uuid.UUID(int=value)
        except ValueError:
            uuid.UUID(hex=value)
        return str(value)


class UUIDFields(ovoo_fields.AutoTypedField):
    AUTO_TYPE = UUID()


class DateTimeField(ovoo_fields.DateTimeField):
    def __init__(self, tzinfo_aware=False, **kwargs):
        super().__init__(tzinfo_aware, **kwargs)


class ObjectField(ovoo_fields.ObjectField):
    pass


class IPV4AddressField(ovoo_fields.IPV4AddressField):

    def coerce(self, obj, attr, value):
        try:
            # make sure that DNS Python agrees that it is a valid IP address
            ipv4.inet_aton(str(value))
        except dns.exception.SyntaxError:
            raise ValueError()
        value = super().coerce(obj, attr, value)
        # we use this field as a string, not need a netaddr.IPAdress
        # as oslo.versionedobjects is using
        return str(value)


class IPV6AddressField(ovoo_fields.IPV6AddressField):

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        # we use this field as a string, not need a netaddr.IPAdress
        # as oslo.versionedobjects is using
        return str(value)


class IPV4AndV6AddressField(ovoo_fields.IPV4AndV6AddressField):

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        # we use this field as a string, not need a netaddr.IPAdress
        # as oslo.versionedobjects is using
        return str(value)


class Enum(ovoo_fields.Enum):
    def get_schema(self):
        return {
            'enum': self._valid_values,
            'type': 'any'
        }


class EnumField(ovoo_fields.BaseEnumField):
    def __init__(self, valid_values, **kwargs):
        self.AUTO_TYPE = Enum(valid_values=valid_values)
        super().__init__(**kwargs)


class DomainField(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        if value is None:
            return
        domain = value.split('.')
        for host in domain:
            if len(host) > 63:
                raise ValueError(f'Host {host} is too long')
        if not value.endswith('.'):
            raise ValueError(f'Domain {value} does not end with a dot')
        if not constants.RE_ZONENAME.match(value):
            raise ValueError(f'Domain {value} is invalid')
        return value


class EmailField(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        if value.count('@') != 1:
            raise ValueError(f'{value} is not an email')
        email = value.replace('@', '.')
        if not constants.RE_ZONENAME.match(f'{email}.'):
            raise ValueError(f'Email {value} is invalid')
        return value


class HostField(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        if value is None:
            return
        hostname = value.split('.')
        for host in hostname:
            if len(host) > 63:
                raise ValueError(f'Host {host} is too long')
        if value.endswith('.') is False:
            raise ValueError(f'Host name {value} does not end with a dot')
        if not constants.RE_HOSTNAME.match(value):
            raise ValueError(f'Host name {value} is invalid')
        return value


class SRVField(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        if value is None:
            return
        srvtype = value.split('.')
        for host in srvtype:
            if len(host) > 63:
                raise ValueError(f'Host {host} is too long')
        if value.endswith('.') is False:
            raise ValueError(f'Host name {value} does not end with a dot')
        if not constants.RE_SRV_HOST_NAME.match(value):
            raise ValueError(f'Host name {value} is not a SRV record')
        return value


class TxtField(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        if value.endswith('\\'):
            raise ValueError("Do NOT put '\\' into end of TXT record")
        return value


class Sshfp(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        if not constants.RE_SSHFP_FINGERPRINT.match(f'{value}'):
            raise ValueError(f'Host name {value} is not a SSHFP record')
        return value


class TldField(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        if not constants.RE_TLDNAME.match(value):
            raise ValueError(f'{value} is not a TLD')
        return value


class NaptrFlagsField(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        if len(value) > 255:
            raise ValueError(
                f'NAPTR record {value} flags field cannot be longer than '
                '255 characters'
            )
        if not constants.RE_NAPTR_FLAGS.match(f'{value}'):
            raise ValueError(
                f'NAPTR record {value} flags can be S, A, U and P'
            )
        return value


class NaptrServiceField(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        if len(value) > 255:
            raise ValueError(
                f'NAPTR record {value} service field cannot be longer than '
                '255 characters'
            )
        if not constants.RE_NAPTR_SERVICE.match(f'{value}'):
            raise ValueError(
                f'{value} NAPTR record service is invalid'
            )
        return value


class NaptrRegexpField(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        if len(value) > 255:
            raise ValueError(
                f'NAPTR record {value} regexp field cannot be longer than '
                '255 characters'
            )
        if value:
            if not constants.RE_NAPTR_REGEXP.match(f'{value}'):
                raise ValueError(f'{value} NAPTR record is invalid')
        return value


class CaaPropertyField(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        prpt = value.split(' ', 1)
        tag = prpt[0]
        val = prpt[1]
        if tag == 'issue' or tag == 'issuewild':
            entries = val.split(';')
            idn = entries.pop(0)
            domain = idn.split('.')
            for host in domain:
                if len(host) > 63:
                    raise ValueError(f'Host {host} is too long')
            idn_with_dot = idn + '.'
            if not constants.RE_ZONENAME.match(idn_with_dot):
                raise ValueError(f'Domain {idn} is invalid')
            for entry in entries:
                if not constants.RE_KVP.match(entry):
                    raise ValueError(f'{entry} is not a valid key-value pair')
        elif tag == 'iodef':
            if constants.RE_URL_MAIL.match(val):
                parts = val.split('@')
                idn = parts[1]
                domain = idn.split('.')
                for host in domain:
                    if len(host) > 63:
                        raise ValueError(f'Host {host} is too long')
                idn_with_dot = idn + '.'
                if not constants.RE_ZONENAME.match(idn_with_dot):
                    raise ValueError(f'Domain {idn} is invalid')
            elif constants.RE_URL_HTTP.match(val):
                parts = val.split('/')
                idn = parts[2]
                domain = idn.split('.')
                for host in domain:
                    if len(host) > 63:
                        raise ValueError(f'Host {host} is too long')
                idn_with_dot = idn + '.'
                if not constants.RE_ZONENAME.match(idn_with_dot):
                    raise ValueError(f'Domain {idn} is invalid')
            else:
                raise ValueError(f'{val} is not a valid URL')
        else:
            raise ValueError(
                f"Property tag {value} must be 'issue', 'issuewild' or 'iodef'"
            )
        return value


class CertTypeField(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        if not constants.RE_CERT_TYPE.match(f'{value}'):
            raise ValueError(
                f'Cert type {value} is not a valid Mnemonic or value'
            )
        return value


class CertAlgoField(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)
        if not constants.RE_CERT_ALGO.match(f'{value}'):
            raise ValueError(
                f'Cert Algo {value} is not a valid Mnemonic or value'
            )
        return value


class Any(ovoo_fields.FieldType):
    @staticmethod
    def coerce(obj, attr, value):
        return value


class AnyField(ovoo_fields.AutoTypedField):
    AUTO_TYPE = Any()


class BaseObject(ovoo_fields.FieldType):
    @staticmethod
    def coerce(obj, attr, value):
        return value


class BaseObjectField(ovoo_fields.AutoTypedField):
    AUTO_TYPE = BaseObject()


class IPOrHost(IPV4AndV6AddressField):
    def __init__(self, nullable=False, read_only=False,
                 default=ovoo_fields.UnspecifiedDefault):
        super().__init__(
            nullable=nullable, default=default, read_only=read_only
        )

    def coerce(self, obj, attr, value):
        try:
            value = super().coerce(obj, attr, value)
        except ValueError:
            if not constants.RE_ZONENAME.match(value):
                raise ValueError(f'{value} is not IP address or host name')
        return value


class DenylistFields(StringFields):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def coerce(self, obj, attr, value):
        value = super().coerce(obj, attr, value)

        if value is None:
            return self._null(obj, attr)

        # determine the validity if a regex expression filter has been used.
        msg = f'{value} is not a valid regular expression'
        if not len(value):
            raise ValueError(msg)
        try:
            re.compile(value)
        except Exception:
            raise ValueError(msg)

        return value



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_log import log as logging

from designate.notification_handler import neutron
from designate.tests import base_fixtures
import designate.tests.functional


LOG = logging.getLogger(__name__)


class NeutronFloatingHandlerTest(designate.tests.functional.TestCase):
    def setUp(self):
        super().setUp()

        zone = self.create_zone()
        self.zone_id = zone['id']
        self.config(zone_id=zone['id'], group='handler:neutron_floatingip')
        formats = ['%(octet0)s-%(octet1)s-%(octet2)s-%(octet3)s.%(zone)s',
                   '%(octet0)s-%(octet1)s-%(octet2)s-%(octet3)s.X.%(zone)s']
        self.config(formatv4=formats, group='handler:neutron_floatingip')

        self.plugin = neutron.NeutronFloatingHandler()

    def test_floatingip_associate(self):
        event_type = 'floatingip.update.end'
        fixture = base_fixtures.get_notification_fixture(
            'neutron', event_type + '_associate'
        )

        self.assertIn(event_type, self.plugin.get_event_types())

        criterion = {'zone_id': self.zone_id}

        # Ensure we start with only SOA and NS records
        records = self.central_service.find_records(self.admin_context,
                                                    criterion)

        self.assertEqual(2, len(records))

        self.plugin.process_notification(
            self.admin_context.to_dict(), event_type, fixture['payload'])

        # Ensure we now have exactly 2 records, plus SOA & NS
        recordsets = self.central_service.find_recordsets(self.admin_context,
                                                          criterion)

        self.assertEqual(4, len(recordsets))

    def test_floatingip_disassociate(self):
        start_event_type = 'floatingip.update.end'
        start_fixture = base_fixtures.get_notification_fixture(
            'neutron', start_event_type + '_associate')
        self.plugin.process_notification(self.admin_context.to_dict(),
                                         start_event_type,
                                         start_fixture['payload'])

        event_type = 'floatingip.update.end'
        fixture = base_fixtures.get_notification_fixture(
            'neutron', event_type + '_disassociate'
        )

        self.assertIn(event_type, self.plugin.get_event_types())

        criterion = {'zone_id': self.zone_id}

        # Ensure we start with exactly 2 records, plus NS and SOA
        records = self.central_service.find_records(self.admin_context,
                                                    criterion)

        self.assertEqual(4, len(records))

        self.plugin.process_notification(
            self.admin_context.to_dict(), event_type, fixture['payload'])

        # Ensure we now have exactly 0 recordsets, plus NS and SOA
        recordsets = self.central_service.find_recordsets(self.admin_context,
                                                          criterion)

        self.assertEqual(2, len(recordsets))

    def test_floatingip_delete(self):
        start_event_type = 'floatingip.update.end'
        start_fixture = base_fixtures.get_notification_fixture(
            'neutron', start_event_type + '_associate'
        )
        self.plugin.process_notification(self.admin_context.to_dict(),
                                         start_event_type,
                                         start_fixture['payload'])

        event_type = 'floatingip.delete.start'
        fixture = base_fixtures.get_notification_fixture(
            'neutron', event_type
        )

        self.assertIn(event_type, self.plugin.get_event_types())

        criterion = {'zone_id': self.zone_id}

        # Ensure we start with exactly 2 records, plus NS and SOA
        records = self.central_service.find_records(self.admin_context,
                                                    criterion)
        self.assertEqual(4, len(records))

        self.plugin.process_notification(
            self.admin_context.to_dict(), event_type, fixture['payload'])

        # Ensure we now have exactly 0 recordsets, plus NS and SOA
        recordsets = self.central_service.find_recordsets(self.admin_context,
                                                          criterion)

        self.assertEqual(2, len(recordsets))



# flake8: noqa
# Copyright (c) <2011>, Jonathan LaCour
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import inspect

import pecan
import pecan.rest
import pecan.routing

from designate import exceptions
from designate.central import rpcapi as central_rpcapi


class RestController(pecan.rest.RestController):
    """
    Extension for Pecan's RestController to better handle POST/PUT/PATCH
    requests.

    Ideally, we get these additions merged upstream.
    """

    # default sort_keys.  The Controllers can override this.
    SORT_KEYS = ['created_at', 'id']

    @property
    def central_api(self):
        return central_rpcapi.CentralAPI.get_instance()

    @classmethod
    def _apply_filter_params(cls, params, accepted_filters, criterion):
        duplicate = []
        invalid = []

        for k in params:
            if isinstance(params[k], list):
                duplicate.append(k)
                continue

            if k in accepted_filters:
                criterion[k] = params[k].replace("*", "%")
            else:
                invalid.append(k)

        if duplicate:
            raise exceptions.BadRequest(
                'Duplicate filters %s' % ', '.join(duplicate))

        if invalid:
            raise exceptions.BadRequest(
                'Invalid filters %s' % ', '.join(invalid))

        return criterion

    def _handle_post(self, method, remainder, request=None):
        """
        Routes ``POST`` actions to the appropriate controller.
        """
        # route to a post_all or get if no additional parts are available
        if not remainder or remainder == ['']:
            controller = self._find_controller('post_all', 'post')
            if controller:
                return controller, []
            pecan.abort(405)

        controller = getattr(self, remainder[0], None)
        if controller and not inspect.ismethod(controller):
            return pecan.routing.lookup_controller(controller, remainder[1:],
                                                   request=request)

        # finally, check for the regular post_one/post requests
        controller = self._find_controller('post_one', 'post')
        if controller:
            return controller, remainder

        pecan.abort(405)

    def _handle_patch(self, method, remainder, request=None):
        """
        Routes ``PATCH`` actions to the appropriate controller.
        """
        # route to a patch_all or get if no additional parts are available
        if not remainder or remainder == ['']:
            controller = self._find_controller('patch_all', 'patch')
            if controller:
                return controller, []
            pecan.abort(405)

        controller = getattr(self, remainder[0], None)
        if controller and not inspect.ismethod(controller):
            return pecan.routing.lookup_controller(controller, remainder[1:],
                                                   request=request)

        # finally, check for the regular patch_one/patch requests
        controller = self._find_controller('patch_one', 'patch')
        if controller:
            return controller, remainder

        pecan.abort(405)

    def _handle_put(self, method, remainder, request=None):
        """
        Routes ``PUT`` actions to the appropriate controller.
        """
        # route to a put_all or get if no additional parts are available
        if not remainder or remainder == ['']:
            controller = self._find_controller('put_all', 'put')
            if controller:
                return controller, []
            pecan.abort(405)

        controller = getattr(self, remainder[0], None)
        if controller and not inspect.ismethod(controller):
            return pecan.routing.lookup_controller(controller, remainder[1:],
                                                   request=request)

        # finally, check for the regular put_one/put requests
        controller = self._find_controller('put_one', 'put')
        if controller:
            return controller, remainder

        pecan.abort(405)

    def _handle_delete(self, method, remainder, request=None):
        """
        Routes ``DELETE`` actions to the appropriate controller.
        """
        # route to a delete_all or get if no additional parts are available
        if not remainder or remainder == ['']:
            controller = self._find_controller('delete_all', 'delete')
            if controller:
                return controller, []
            pecan.abort(405)

        controller = getattr(self, remainder[0], None)
        if controller and not inspect.ismethod(controller):
            return pecan.routing.lookup_controller(controller, remainder[1:],
                                                   request=request)

        # finally, check for the regular delete_one/delete requests
        controller = self._find_controller('delete_one', 'delete')
        if controller:
            return controller, remainder

        pecan.abort(405)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import dns.message

import designate.conf
from designate import dnsmiddleware
from designate import exceptions
import oslotest.base

CONF = designate.conf.CONF


@mock.patch('designate.context.DesignateContext.get_admin_context',
            mock.Mock())
class TestSerializationMiddleware(oslotest.base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.message = dns.message.Message()
        self.application = mock.Mock()
        self.application.return_value = [None, self.message]
        self.dns_middleware = dnsmiddleware.SerializationMiddleware(
            self.application
        )

    @mock.patch.object(dns.message, 'from_wire')
    def test_unknown_tsig_key(self, mock_from_wire):
        mock_from_wire.side_effect = dns.message.UnknownTSIGKey()
        next(self.dns_middleware(
            {'payload': 'payload', 'addr': ['192.0.2.1', 5353]})
        )

    @mock.patch.object(dns.message, 'from_wire')
    def test_bad_signature(self, mock_from_wire):
        mock_from_wire.side_effect = dns.tsig.BadSignature()
        next(self.dns_middleware(
            {'payload': 'payload', 'addr': ['192.0.2.1', 5353]})
        )

    @mock.patch.object(dns.message, 'from_wire')
    def test_dns_exception(self, mock_from_wire):
        mock_from_wire.side_effect = dns.exception.DNSException()
        next(self.dns_middleware(
            {'payload': 'payload', 'addr': ['192.0.2.1', 5353]})
        )

    @mock.patch.object(dns.message, 'from_wire')
    def test_general_exception(self, mock_from_wire):
        mock_from_wire.side_effect = Exception()
        next(self.dns_middleware(
            {'payload': 'payload', 'addr': ['192.0.2.1', 5353]})
        )

    @mock.patch.object(dns.message, 'from_wire', mock.Mock())
    def test_message_to_wire(self):
        self.assertEqual(
            self.message.to_wire(max_size=65535),
            next(self.dns_middleware(
                {'payload': 'payload', 'addr': ['192.0.2.1', 5353]})
            ))

    @mock.patch.object(dns.message, 'from_wire', mock.Mock())
    def test_renderer_to_wire(self):
        self.message = dns.renderer.Renderer()
        self.application.return_value = [self.message]

        self.assertEqual(
            self.message.get_wire(),
            next(self.dns_middleware(
                {'payload': 'payload', 'addr': ['192.0.2.1', 5353]})
            ))


@mock.patch('designate.context.get_current', mock.Mock())
class TestTsigInfoMiddleware(oslotest.base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.application = mock.Mock()
        self.storage = mock.Mock()
        self.dns_middleware = dnsmiddleware.TsigInfoMiddleware(
            self.application, self.storage
        )

    def test_process_request(self):
        mock_context = mock.Mock()
        mock_request = mock.Mock()
        mock_request.keyname.to_text.return_value = 'test'
        mock_request.environ = {'context': mock_context}

        self.assertIsNone(self.dns_middleware.process_request(mock_request))

        self.storage.find_tsigkey.assert_called_with(
            mock.ANY, {'name': 'test'}
        )

    def test_process_request_tsig_key_not_found(self):
        mock_context = mock.Mock()
        mock_request = mock.Mock()
        mock_request.environ = {'context': mock_context}

        self.storage.find_tsigkey.side_effect = exceptions.TsigKeyNotFound()
        self.dns_middleware._build_error_response = mock.Mock()

        self.assertEqual(
            self.dns_middleware._build_error_response(),
            self.dns_middleware.process_request(mock_request)
        )


class TestDNSMiddleware(oslotest.base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.application = mock.Mock(name='application')
        self.dns_application = dnsmiddleware.DNSMiddleware(self.application)

    @mock.patch.object(dnsmiddleware.DNSMiddleware, 'process_request')
    def test_call(self, mock_process_request):
        request = mock.Mock()
        self.dns_application(request)

        mock_process_request.assert_called_with(request)

    @mock.patch.object(dnsmiddleware.DNSMiddleware, 'process_response')
    @mock.patch.object(dnsmiddleware.DNSMiddleware, 'process_request')
    def test_call_with_none(self, mock_process_request, mock_process_response):
        mock_process_request.return_value = None

        self.dns_application(None)

        mock_process_request.assert_called_with(None)
        mock_process_response.assert_called_with(self.application())

    def test_process_request(self):
        self.assertIsNone(self.dns_application.process_request(mock.Mock()))

    def test_process_response(self):
        response = mock.Mock()
        self.assertEqual(
            response, self.dns_application.process_response(response)
        )

    def test_build_error_response(self):
        self.assertIsInstance(
            self.dns_application._build_error_response(), dns.message.Message
        )



# Copyright 2012 Managed I.T.
# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_log import log as logging
import oslo_messaging as messaging

import designate.conf
from designate import notification_handler
from designate import rpc
from designate import service


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class Service(service.Service):
    def __init__(self):
        super().__init__(
            self.service_name, threads=CONF['service:sink'].threads
        )

        # Initialize extensions
        self._notification_listener = None
        self.handlers = self.init_extensions()
        self.allowed_event_types = self.get_allowed_event_types(self.handlers)

    @property
    def service_name(self):
        return 'sink'

    @staticmethod
    def init_extensions():
        """Loads and prepares all enabled extensions"""
        notification_handlers = notification_handler.get_notification_handlers(
            CONF['service:sink'].enabled_notification_handlers
        )

        if not notification_handlers:
            LOG.warning('No designate-sink handlers enabled or loaded')

        return notification_handlers

    @staticmethod
    def get_allowed_event_types(handlers):
        """Build a list of all allowed event types."""
        allowed_event_types = []

        for handler in handlers:
            for event_type in handler.get_event_types():
                if event_type in allowed_event_types:
                    continue
                allowed_event_types.append(event_type)

        return allowed_event_types

    def start(self):
        super().start()

        # Setup notification subscriptions and start consuming
        targets = self._get_targets()

        # TODO(ekarlso): Change this is to endpoint objects rather then
        # ourselves?
        if targets:
            self._notification_listener = rpc.get_notification_listener(
                targets, [self],
                pool=CONF['service:sink'].listener_pool_name
            )
            self._notification_listener.start()

    def stop(self, graceful=True):
        if self._notification_listener:
            self._notification_listener.stop()
        super().stop(graceful)

    def _get_targets(self):
        """
        Set's up subscriptions for the various exchange+topic combinations that
        we have a handler for.
        """
        targets = []
        for handler in self.handlers:
            exchange, topics = handler.get_exchange_topics()

            for topic in topics:
                target = messaging.Target(exchange=exchange, topic=topic)
                targets.append(target)
        return targets

    def info(self, context, publisher_id, event_type, payload, metadata):
        """
        Processes an incoming notification, offering each extension the
        opportunity to handle it.
        """
        if event_type not in self.allowed_event_types:
            return

        for handler in self.handlers:
            if event_type in handler.get_event_types():
                LOG.debug('Found handler for: %s', event_type)
                handler.process_notification(context, event_type, payload)



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from designate import objects
from designate.objects.adapters.api_v2 import base


class QuotaAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.Quota
    MODIFICATIONS = {
        'fields': {
            'resource': {
                'read_only': False
            },
            'hard_limit': {
                'read_only': False
            },
        },
        'options': {
            'links': True,
            'resource_name': 'quota',
            'collection_name': 'quotas',
        }
    }


class QuotaListAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.QuotaList
    MODIFICATIONS = {
        'options': {
            'links': True,
            'resource_name': 'quota',
            'collection_name': 'quotas',
        }
    }

    @classmethod
    def render_list(cls, list_objects, *args, **kwargs):
        r_list = {}
        for obj in list_objects:
            r_list[obj.resource] = obj.hard_limit
        return r_list

    @classmethod
    def parse_list(cls, values, output_object, *args, **kwargs):
        for key, value in values.items():
            # Add the object to the list
            output_object.append(
                cls.ADAPTER_OBJECT.LIST_ITEM_TYPE.from_dict(
                    {
                        'resource': key,
                        'hard_limit': value,
                    }
                )
            )
        return output_object



#    Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""unique_service_status

Revision ID: 93a00a815f07
Revises: 7977deaa5167
Create Date: 2022-08-01 17:17:44.572964

"""
import sys

from alembic import op
from oslo_log import log as logging
import sqlalchemy as sa

from designate.storage.sqlalchemy.alembic import legacy_utils

# revision identifiers, used by Alembic.
revision = '93a00a815f07'
down_revision = '7977deaa5167'
branch_labels = None
depends_on = None

LOG = logging.getLogger()

# Equivalent to legacy sqlalchemy-migrate revision 100_unique_service_status

EXPLANATION = """
You need to manually remove duplicate entries from the database.

The error message was:
%s
"""


def upgrade() -> None:
    # Check if the equivalent legacy migration has already run
    if not legacy_utils.is_migration_needed(100):
        return

    try:
        with op.batch_alter_table('service_statuses') as batch_op:
            batch_op.create_unique_constraint('unique_service_status',
                                              ['service_name', 'hostname'])
    except sa.exc.IntegrityError as e:
        LOG.error(EXPLANATION, e)
        # Use sys.exit so we don't blow up with a huge trace
        sys.exit(1)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_log import log as logging

import designate.conf
from designate import objects
from designate.tests.functional.api import v2

CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class ZoneExportsTest(v2.ApiV2TestCase):
    def setUp(self):
        super().setUp()
        self.zone = self.create_zone()

    def create_zone_export(self, location='designate://'):
        return self.storage.create_zone_export(
            self.admin_context,
            objects.ZoneExport(
                status='ACTIVE',
                task_type='EXPORT',
                location=location,
                zone_id=self.zone['id']
            )
        )

    def test_export(self):
        zone_export = self.create_zone_export()
        response = self.client.get(
            '/zones/tasks/exports/%s/export' % zone_export['id'],
            headers={'X-Test-Role': 'member'}
        )

        self.assertEqual(200, response.status_int)
        self.assertEqual('text/dns', response.content_type)

    def test_export_cannot_be_exported_synchronously(self):
        zone_export = self.create_zone_export(location=None)

        self._assert_exception(
            'bad_request', 400, self.client.get,
            '/zones/tasks/exports/%s/export' % zone_export['id'],
            headers={'X-Test-Role': 'member'}
        )



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_log import versionutils
from oslo_policy import policy

from designate.common import constants
from designate.common.policies import base

DEPRECATED_REASON = """
The top-level domain API now supports system scope and default roles.
"""

deprecated_create_tld = policy.DeprecatedRule(
    name="create_tld",
    check_str=base.RULE_ADMIN,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_find_tlds = policy.DeprecatedRule(
    name="find_tlds",
    check_str=base.RULE_ADMIN,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_get_tld = policy.DeprecatedRule(
    name="get_tld",
    check_str=base.RULE_ADMIN,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_update_tld = policy.DeprecatedRule(
    name="update_tld",
    check_str=base.RULE_ADMIN,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_delete_tld = policy.DeprecatedRule(
    name="delete_tld",
    check_str=base.RULE_ADMIN,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)


rules = [
    policy.DocumentedRuleDefault(
        name="create_tld",
        check_str=base.SYSTEM_ADMIN,
        scope_types=[constants.PROJECT],
        description="Create Tld",
        operations=[
            {
                'path': '/v2/tlds',
                'method': 'POST'
            }
        ],
        deprecated_rule=deprecated_create_tld
    ),
    policy.DocumentedRuleDefault(
        name="find_tlds",
        check_str=base.SYSTEM_READER,
        scope_types=[constants.PROJECT],
        description="List Tlds",
        operations=[
            {
                'path': '/v2/tlds',
                'method': 'GET'
            }
        ],
        deprecated_rule=deprecated_find_tlds
    ),
    policy.DocumentedRuleDefault(
        name="get_tld",
        check_str=base.SYSTEM_READER,
        scope_types=[constants.PROJECT],
        description="Show Tld",
        operations=[
            {
                'path': '/v2/tlds/{tld_id}',
                'method': 'GET'
            }
        ],
        deprecated_rule=deprecated_get_tld
    ),
    policy.DocumentedRuleDefault(
        name="update_tld",
        check_str=base.SYSTEM_ADMIN,
        scope_types=[constants.PROJECT],
        description="Update Tld",
        operations=[
            {
                'path': '/v2/tlds/{tld_id}',
                'method': 'PATCH'
            }
        ],
        deprecated_rule=deprecated_update_tld
    ),
    policy.DocumentedRuleDefault(
        name="delete_tld",
        check_str=base.SYSTEM_ADMIN,
        scope_types=[constants.PROJECT],
        description="Delete Tld",
        operations=[
            {
                'path': '/v2/tlds/{tld_id}',
                'method': 'DELETE'
            }
        ],
        deprecated_rule=deprecated_delete_tld
    )
]


def list_rules():
    return rules



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import patch

import oslo_messaging as messaging

from designate.central import service as central_service
import designate.conf
from designate import exceptions
from designate.tests.functional.api import v2


CONF = designate.conf.CONF


class ApiV2TsigKeysTest(v2.ApiV2TestCase):
    def setUp(self):
        super().setUp()

        # Set the policy to accept everyone as an admin, as this is an
        # admin-only API
        self.policy({'admin': '@'})

    def test_create_tsigkey(self):
        # Create a TSIG Key
        fixture = self.get_tsigkey_fixture(0)
        response = self.client.post_json('/tsigkeys/', fixture)

        # Check the headers are what we expect
        self.assertEqual(201, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the generated values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIn('created_at', response.json)
        self.assertIsNone(response.json['updated_at'])

        # Check the supplied values returned are what we expect
        self.assertLessEqual(fixture.items(), response.json.items())

    def test_create_tsigkey_validation(self):
        # NOTE: The schemas should be tested separately to the API. So we
        #       don't need to test every variation via the API itself.
        # Fetch a fixture
        fixture = self.get_tsigkey_fixture(0)

        # Add a junk field to the body
        fixture['junk'] = 'Junk Field'

        # Ensure it fails with a 400
        body = fixture

        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/tsigkeys', body)

    def test_create_tsigkey_name_missing(self):
        fixture = self.get_tsigkey_fixture(0)
        del fixture['name']
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/tsigkeys', body)

    def test_create_tsigkey_name_too_long(self):
        fixture = self.get_tsigkey_fixture(0)
        fixture['name'] = 'test-key-' + 'x' * 160
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/tsigkeys', body)

    def test_create_tsigkey_algorithm_missing(self):
        fixture = self.get_tsigkey_fixture(0)
        del fixture['algorithm']
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/tsigkeys', body)

    def test_create_tsigkey_algorithm_invalid_type(self):
        fixture = self.get_tsigkey_fixture(0)
        fixture['algorithm'] = "ABC"
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/tsigkeys', body)

    def test_create_tsigkey_secret_missing(self):
        fixture = self.get_tsigkey_fixture(0)
        del fixture['secret']
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/tsigkeys', body)

    def test_create_tsigkey_empty_secret(self):
        CONF.set_override(
            'allow_empty_secrets_for_tsig',
            False,
            'service:api'
        )
        fixture = self.get_tsigkey_fixture(0)
        fixture['secret'] = ''
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/tsigkeys', body)

    def test_create_tsigkey_secret_too_long(self):
        fixture = self.get_tsigkey_fixture(0)
        fixture['secret'] = 'x' * 161
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/tsigkeys', body)

    def test_create_tsigkey_scope_missing(self):
        fixture = self.get_tsigkey_fixture(0)
        del fixture['scope']
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/tsigkeys', body)

    def test_create_tsigkey_scope_invalid_type(self):
        fixture = self.get_tsigkey_fixture(0)
        fixture['scope'] = "ABC"
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/tsigkeys', body)

    def test_create_tsigkey_resource_id_missing(self):
        fixture = self.get_tsigkey_fixture(0)
        del fixture['resource_id']
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/tsigkeys', body)

    def test_create_tsigkey_invalid_resource_id(self):
        fixture = self.get_tsigkey_fixture(0)
        fixture['resource_id'] = "xyz"
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/tsigkeys', body)

    def test_create_tsigkey_duplicate(self):
        # Prepare a TSIG Key fixture
        fixture = self.get_tsigkey_fixture(0)
        body = fixture

        # Create the first TSIG Key
        response = self.client.post_json('/tsigkeys', body)
        self.assertEqual(201, response.status_int)

        self._assert_exception('duplicate_tsigkey', 409,
                               self.client.post_json, '/tsigkeys', body)

    def test_get_tsigkeys(self):
        response = self.client.get('/tsigkeys/')

        # Check the headers are what we expect
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('tsigkeys', response.json)
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # We should start with 0 tsigkeys
        self.assertEqual(0, len(response.json['tsigkeys']))

        data = [self.create_tsigkey(name='tsigkey-%s' % i)
                for i in range(1, 10)]
        self._assert_paging(data, '/tsigkeys', key='tsigkeys')
        self._assert_invalid_paging(data, '/tsigkeys', key='tsigkeys')

    @patch.object(central_service.Service, 'find_tsigkeys',
                  side_effect=messaging.MessagingTimeout())
    def test_get_tsigkeys_timeout(self, _):
        self._assert_exception('timeout', 504, self.client.get, '/tsigkeys/')

    def test_get_tsigkey(self):
        # Create a tsigkey
        tsigkey = self.create_tsigkey()

        response = self.client.get('/tsigkeys/%s' % tsigkey.id,
                                   headers=[('Accept', 'application/json')])

        # Check the headers are what we expect
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the generated values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIn('created_at', response.json)
        self.assertIsNone(response.json['updated_at'])

        # Check the supplied values returned are what we expect
        self.assertEqual(tsigkey.name, response.json['name'])
        self.assertEqual(
            tsigkey.algorithm, response.json['algorithm'])
        self.assertEqual(tsigkey.secret, response.json['secret'])
        self.assertEqual(tsigkey.scope, response.json['scope'])
        self.assertEqual(
            tsigkey.resource_id, response.json['resource_id'])

    def test_get_tsigkey_invalid_id(self):
        self._assert_invalid_uuid(self.client.get, '/tsigkeys/%s')

    @patch.object(central_service.Service, 'get_tsigkey',
                  side_effect=messaging.MessagingTimeout())
    def test_get_tsigkey_timeout(self, _):
        url = '/tsigkeys/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'
        self._assert_exception('timeout', 504, self.client.get, url,
                               headers={'Accept': 'application/json'})

    @patch.object(central_service.Service, 'get_tsigkey',
                  side_effect=exceptions.TsigKeyNotFound())
    def test_get_tsigkey_missing(self, _):
        url = '/tsigkeys/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'
        self._assert_exception('tsigkey_not_found', 404, self.client.get, url,
                               headers={'Accept': 'application/json'})

    def test_update_tsigkey(self):
        # Create a TSIG Key
        tsigkey = self.create_tsigkey()

        # Prepare an update body
        body = {'secret': 'prefix-%s' % tsigkey.secret}

        response = self.client.patch_json('/tsigkeys/%s' % tsigkey.id, body)

        # Check the headers are what we expect
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIsNotNone(response.json['updated_at'])
        self.assertEqual('prefix-%s' % tsigkey['secret'],
                         response.json['secret'])

    def test_update_tsigkey_json_patch_not_implemented(self):
        self._assert_exception(
            'unknown', 500, self.client.patch_json,
            '/tsigkeys/b0f331f1-7683-4444-aa68-b1d29e40924b',
            {'secret': 'secret'},
            content_type='application/json-patch+json'
        )

    def test_update_tsigkey_invalid_id(self):
        self._assert_invalid_uuid(self.client.patch_json, '/tsigkeys/%s')

    @patch.object(central_service.Service, 'get_tsigkey',
                  side_effect=exceptions.DuplicateTsigKey())
    def test_update_tsigkey_duplicate(self, _):
        # Prepare an update body
        body = {'name': 'AnyOldName'}

        url = '/tsigkeys/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'

        # Ensure it fails with a 409
        self._assert_exception('duplicate_tsigkey', 409,
                               self.client.patch_json, url, body)

    def test_update_tsigkey_secret_too_long(self):
        tsigkey = self.create_tsigkey()
        body = {'secret': 'x' * 161}
        url = '/tsigkeys/%s' % tsigkey.id
        self._assert_exception('invalid_object', 400,
                               self.client.patch_json, url, body)

    def test_update_tsigkey_invalid_scope_type(self):
        tsigkey = self.create_tsigkey()
        body = {'scope': 'abc'}
        url = '/tsigkeys/%s' % tsigkey.id
        self._assert_exception('invalid_object', 400,
                               self.client.patch_json, url, body)

    def test_update_tsigkey_invalid_algorithm(self):
        tsigkey = self.create_tsigkey()
        body = {'algorithm': 'abc'}
        url = '/tsigkeys/%s' % tsigkey.id
        self._assert_exception('invalid_object', 400,
                               self.client.patch_json, url, body)

    def test_update_tsigkey_junk_field(self):
        tsigkey = self.create_tsigkey()
        body = {'junk': 'abc'}
        url = '/tsigkeys/%s' % tsigkey.id
        self._assert_exception('invalid_object', 400,
                               self.client.patch_json, url, body)

    @patch.object(central_service.Service, 'get_tsigkey',
                  side_effect=messaging.MessagingTimeout())
    def test_update_tsigkey_timeout(self, _):
        # Prepare an update body
        body = {'name': 'AnyOldName'}

        url = '/tsigkeys/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'

        # Ensure it fails with a 504
        self._assert_exception('timeout', 504, self.client.patch_json,
                               url, body)

    @patch.object(central_service.Service, 'get_tsigkey',
                  side_effect=exceptions.TsigKeyNotFound())
    def test_update_tsigkey_missing(self, _):
        # Prepare an update body
        body = {'name': 'AnyOldName'}

        url = '/tsigkeys/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'

        # Ensure it fails with a 404
        self._assert_exception('tsigkey_not_found', 404,
                               self.client.patch_json, url, body)

    def test_delete_tsigkey(self):
        tsigkey = self.create_tsigkey()

        self.client.delete('/tsigkeys/%s' % tsigkey['id'], status=204)

    def test_delete_tsigkey_invalid_id(self):
        self._assert_invalid_uuid(self.client.delete, '/tsigkeys/%s')

    @patch.object(central_service.Service, 'delete_tsigkey',
                  side_effect=messaging.MessagingTimeout())
    def test_delete_tsigkey_timeout(self, _):
        url = '/tsigkeys/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'

        self._assert_exception('timeout', 504, self.client.delete, url)

    @patch.object(central_service.Service, 'delete_tsigkey',
                  side_effect=exceptions.TsigKeyNotFound())
    def test_delete_tsigkey_missing(self, _):
        url = '/tsigkeys/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'

        self._assert_exception('tsigkey_not_found', 404, self.client.delete,
                               url)



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import jsonschema
from oslo_log import log as logging

from designate import exceptions
from designate import utils

LOG = logging.getLogger(__name__)


class Schema:
    def __init__(self, version, name):
        self.raw_schema = utils.load_schema(version, name)
        self.validator = jsonschema.Draft4Validator(self.raw_schema)

    def validate(self, obj):
        LOG.debug('Validating values: %r', obj)
        errors = []

        for error in self.validator.iter_errors(obj):
            errors.append({
                'path': '.'.join([str(x) for x in error.path]),
                'message': error.message,
                'validator': error.validator
            })

        if errors:
            LOG.debug('Errors in validation: %r', errors)
            raise exceptions.InvalidObject(
                'Provided object does not match schema',
                errors=errors
            )



# Copyright 2020 Cloudification GmbH. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from designate.tests.functional.api import v2


class ApiV2SharedZonesTest(v2.ApiV2TestCase):
    def setUp(self):
        super().setUp()

        self.zone = self.create_zone()
        self.target_project_id = '2'
        self.endpoint_url = '/zones/{}/shares'

    def _create_valid_shared_zone(self):
        return self.client.post_json(
            self.endpoint_url.format(self.zone.id),
            {
                'target_project_id': self.target_project_id,
            }, headers={'X-Test-Role': 'member'}
        )

    def test_share_zone(self):
        response = self._create_valid_shared_zone()

        # Check the headers are what we expect
        self.assertEqual(201, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIn('created_at', response.json)
        self.assertEqual(
            self.target_project_id,
            response.json['target_project_id'])
        self.assertEqual(
            self.zone.id,
            response.json['zone_id'])
        self.assertIsNone(response.json['updated_at'])

    def test_share_zone_with_no_target_id_no_zone_id(self):
        self._assert_exception(
            'invalid_uuid', 400, self.client.post_json,
            self.endpoint_url.format(""), {"target_project_id": ""},
            headers={'X-Test-Role': 'member'}
        )

    def test_share_zone_with_target_id_no_zone_id(self):
        self._assert_exception(
            'invalid_uuid', 400, self.client.post_json,
            self.endpoint_url.format(""), {"target_project_id": "2"},
            headers={'X-Test-Role': 'member'}
        )

    def test_share_zone_with_invalid_zone_id(self):
        self._assert_exception(
            'invalid_uuid', 400, self.client.post_json,
            self.endpoint_url.format("invalid"), {"target_project_id": "2"},
            headers={'X-Test-Role': 'member'}
        )

    def test_get_zone_share(self):
        shared_zone = self._create_valid_shared_zone()

        response = self.client.get(
            '{}/{}'.format(self.endpoint_url.format(self.zone.id),
                           shared_zone.json['id']),
            headers={'X-Test-Role': 'member'}
        )

        # Check the headers are what we expect
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIn('created_at', response.json)
        self.assertEqual(
            self.target_project_id,
            response.json['target_project_id'])
        self.assertEqual(
            self.zone.id,
            response.json['zone_id'])
        self.assertIn('updated_at', response.json)

    def test_list_zone_shares(self):
        response = self.client.get(self.endpoint_url.format(self.zone.id),
                                   headers={'X-Test-Role': 'member'})

        # Check the headers are what we expect
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('shared_zones', response.json)
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # We should start with 0 zone shars
        self.assertEqual(0, len(response.json['shared_zones']))

        self._create_valid_shared_zone()

        data = self.client.get(self.endpoint_url.format(self.zone.id),
                               headers={'X-Test-Role': 'member'})

        self.assertEqual(1, len(data.json['shared_zones']))

    def test_delete_zone_share(self):
        shared_zone = self._create_valid_shared_zone()

        response = self.client.delete(
            '{}/{}'.format(self.endpoint_url.format(self.zone.id),
                           shared_zone.json['id']),
            headers={'X-Test-Role': 'member'}
        )

        # Check the headers are what we expect
        self.assertEqual(204, response.status_int)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_config import fixture as cfg_fixture
from oslo_log import log as logging
import oslotest.base

import designate.conf
from designate import objects


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class FloatingIpTest(oslotest.base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(cfg_fixture.Config(CONF))

    def test_allow_floating_ip_ttl_zero(self):
        floating_ip = objects.FloatingIP(
            ptrdname='ptr1.example.org.',
            description='test',
            address='192.0.2.50',
            ttl=0,
        )
        floating_ip.validate()

        self.assertEqual('ptr1.example.org.', floating_ip.ptrdname)
        self.assertEqual('test', floating_ip.description)
        self.assertEqual('192.0.2.50', floating_ip.address)
        self.assertEqual(0, floating_ip.ttl)



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

from oslo_config import fixture as cfg_fixture
from oslo_log import log as logging
import oslotest.base

import designate.conf
from designate import exceptions
from designate import objects
from designate.objects import adapters
from designate.objects.adapters.api_v2 import base


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class APIv2AdapterTest(oslotest.base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(cfg_fixture.Config(CONF))

    def test_get_base_url(self):
        CONF.set_override('enable_host_header', False, 'service:api')
        CONF.set_override(
            'api_base_uri', 'http://192.0.2.1:9001/', 'service:api'
        )

        mock_request = mock.Mock()
        mock_request.GET = {'foo': 'bar'}
        mock_request.host_url = 'http://192.0.2.2'
        mock_request.path = '/v2/zones'

        base_url = base.APIv2Adapter._get_base_url(mock_request)
        self.assertEqual('http://192.0.2.1:9001', base_url)

    def test_get_base_url_enable_host_header(self):
        CONF.set_override('enable_host_header', True, 'service:api')

        mock_request = mock.Mock()
        mock_request.GET = {'foo': 'bar'}
        mock_request.host_url = 'http://192.0.2.1'
        mock_request.path = '/v2/zones'

        base_url = base.APIv2Adapter._get_base_url(mock_request)
        self.assertEqual('http://192.0.2.1', base_url)


class RecordAPIv2AdapterTest(oslotest.base.BaseTestCase):
    def test_parse_object(self):
        values = '192.0.2.1'

        parsed_object = adapters.RecordAPIv2Adapter.parse_object(
            values, objects.Record()
        )

        self.assertEqual('192.0.2.1', parsed_object.data)

    def test_render_object(self):
        values = '192.0.2.1'

        parsed_object = adapters.RecordAPIv2Adapter.parse_object(
            values, objects.Record()
        )

        self.assertEqual('192.0.2.1', parsed_object.data)

        render_object = adapters.RecordAPIv2Adapter.render_object(
            parsed_object
        )

        self.assertEqual(values, render_object)


class RecordSetAPIv2AdapterTest(oslotest.base.BaseTestCase):
    def test_parse_object(self):
        values = {'name': 'example.org.'}

        parsed_object = adapters.RecordSetAPIv2Adapter.parse_object(
            values, objects.RecordSet()
        )

        self.assertEqual(values['name'], parsed_object.name)

    def test_parse_object_invalid_records(self):
        values = {'name': 'example.org.', 'records': {'foo': 'bar'}}

        self.assertRaisesRegex(
            exceptions.InvalidObject,
            'Provided object does not match schema',
            adapters.RecordSetAPIv2Adapter.parse_object, values,
            objects.RecordSet()
        )

    def test_render_object(self):
        values = {'name': 'example.org.', 'records': []}

        parsed_object = adapters.RecordSetAPIv2Adapter.parse_object(
            values, objects.RecordSet()
        )

        self.assertEqual(values['name'], parsed_object.name)

        render_object = adapters.RecordSetAPIv2Adapter.render_object(
            parsed_object
        )

        self.assertIsInstance(render_object, dict)
        self.assertEqual(values['name'], parsed_object.name)

    def test_get_path(self):
        mock_request = mock.Mock()
        mock_request.path = '/v2/recordsets'
        mock_recordset = mock.Mock()
        mock_recordset.zone_id = 'a-b-c-d'
        expected_path = '/v2/zones/a-b-c-d/recordsets'

        path = adapters.RecordSetAPIv2Adapter._get_path(
            mock_request, mock_recordset
        )
        self.assertEqual(expected_path, path)


class SharedZoneAPIv2AdapterTest(oslotest.base.BaseTestCase):
    def test_render_object(self):
        values = {
            'target_project_id': '60b1dd98-49a7-403b-901a-291ca555da56',
        }

        parsed_object = adapters.SharedZoneAPIv2Adapter.parse_object(
            values, objects.SharedZone()
        )

        render_object = adapters.SharedZoneAPIv2Adapter.render_object(
            parsed_object,
        )

        self.assertIsInstance(render_object, dict)
        self.assertEqual(
            '60b1dd98-49a7-403b-901a-291ca555da56',
            render_object['target_project_id']
        )

    def test_render_object_links(self):
        share_id = '9edbb487-3ca1-4e9c-9503-6dd08a04b8c2'
        zone_id = '9207ed7d-2094-4622-bce9-1af1886e958a'

        mock_request = mock.Mock()
        mock_request.host_url = 'http://192.0.2.1'
        mock_request.path = '/v2/zones/{zone_id}/shares/{zone_share_id}'
        values = {
            'target_project_id': '60b1dd98-49a7-403b-901a-291ca555da56',
        }

        parsed_object = adapters.SharedZoneAPIv2Adapter.parse_object(
            values, objects.SharedZone()
        )
        parsed_object.id = share_id
        parsed_object.zone_id = zone_id

        render_object = adapters.SharedZoneAPIv2Adapter.render_object(
            parsed_object, request=mock_request,
        )

        self.assertEqual(
            f'{mock_request.host_url}/v2/zones/{zone_id}/shares/{share_id}',
            render_object['links']['self']
        )
        self.assertEqual(
            f'{mock_request.host_url}/v2/zones/{zone_id}',
            render_object['links']['zone']
        )


class SharedZoneListAPIv2AdapterTest(oslotest.base.BaseTestCase):
    def test_get_collection_href(self):
        mock_request = mock.Mock()
        mock_request.GET = {'foo': 'bar'}
        mock_request.host_url = 'http://192.0.2.1'
        mock_request.path = '/v2/zones/{zone_id}/shares'

        self.assertEqual(
            'http://192.0.2.1/v2/zones/{zone_id}/shares?foo=bar',
            adapters.SharedZoneListAPIv2Adapter._get_collection_href(
                mock_request, {}
            )
        )


class ZoneAPIv2AdapterTest(oslotest.base.BaseTestCase):
    def test_parse_object(self):
        values = {
            'masters': ['192.0.2.1'],
            'attributes': {'foo': 'bar'},
        }

        adapter = adapters.ZoneAPIv2Adapter.parse_object(
            values, objects.Zone()
        )

        self.assertFalse(values)
        self.assertEqual(1, len(adapter.attributes))
        self.assertEqual('foo', adapter.attributes[0].key)
        self.assertEqual('bar', adapter.attributes[0].value)


class ZoneAttributeAPIv2AdapterTest(oslotest.base.BaseTestCase):
    def test_parse_object(self):
        values = {
            'foo': 'bar'
        }

        parsed_object = adapters.ZoneAttributeAPIv2Adapter.parse_object(
            values, objects.ZoneAttribute()
        )

        self.assertEqual('foo', parsed_object.key)
        self.assertEqual('bar', parsed_object.value)

    def test_render_object(self):
        values = {
            'foo': 'bar'
        }

        parsed_object = adapters.ZoneAttributeAPIv2Adapter.parse_object(
            values, objects.ZoneAttribute()
        )

        self.assertEqual('foo', parsed_object.key)
        self.assertEqual('bar', parsed_object.value)

        render_object = adapters.ZoneAttributeAPIv2Adapter.render_object(
            parsed_object
        )

        self.assertIsInstance(render_object, dict)
        self.assertEqual(values, render_object)


class ZoneAttributeListAPIv2AdapterTest(oslotest.base.BaseTestCase):
    def test_parse_list(self):
        zone_attribute = objects.ZoneAttribute(key='foo', value='bar')

        parsed_object = adapters.ZoneAttributeListAPIv2Adapter.parse_object(
            {}, objects.ZoneAttributeList(objects=[zone_attribute])
        )

        self.assertEqual(1, len(parsed_object))

    def test_render_list(self):
        zone_attribute = objects.ZoneAttribute(key='foo', value='bar')

        parsed_object = adapters.ZoneAttributeListAPIv2Adapter.parse_object(
            {}, objects.ZoneAttributeList(objects=[zone_attribute])
        )

        self.assertEqual(1, len(parsed_object))

        render_object = adapters.ZoneAttributeListAPIv2Adapter.render_list(
            parsed_object
        )

        self.assertEqual({'foo': 'bar'}, render_object)


class ZoneExportAPIv2AdapterTest(oslotest.base.BaseTestCase):
    def test_render_object_links(self):
        export_id = '51790877-ddb7-4351-b3af-23fb6e53098b'

        mock_request = mock.Mock()
        mock_request.GET = {'foo': 'bar'}
        mock_request.host_url = 'designate://192.0.2.1'
        mock_request.path = '/v2/zones/{zone_id}/tasks/export'

        parsed_object = adapters.ZoneExportAPIv2Adapter.parse_object(
            {}, objects.ZoneExport()
        )
        parsed_object.location = (
            'designate://192.0.2.1/v2/zones/{zone_id}/tasks/export'
        )
        parsed_object.id = export_id
        render_object = adapters.ZoneExportAPIv2Adapter.render_object(
            parsed_object, request=mock_request,
        )

        self.assertEqual(
            'designate://192.0.2.1/192.0.2.1/v2/zones/{zone_id}/tasks/export',
            render_object['links']['export']
        )
        self.assertEqual(
            f'designate://192.0.2.1/v2/zones/tasks/exports/{export_id}',
            render_object['links']['self']
        )


class ZoneMasterAPIv2AdapterTest(oslotest.base.BaseTestCase):
    def test_parse_object(self):
        value = '192.0.2.1:5353'

        parsed_object = adapters.ZoneMasterAPIv2Adapter.parse_object(
            value, objects.ZoneMaster()
        )

        self.assertEqual('192.0.2.1', parsed_object.host)
        self.assertEqual(5353, parsed_object.port)

    def test_render_object(self):
        value = '192.0.2.1:5353'

        parsed_object = adapters.ZoneMasterAPIv2Adapter.parse_object(
            value, objects.ZoneMaster()
        )

        self.assertEqual('192.0.2.1', parsed_object.host)
        self.assertEqual(5353, parsed_object.port)

        render_object = adapters.ZoneMasterAPIv2Adapter.render_object(
            parsed_object
        )

        self.assertEqual('192.0.2.1:5353', render_object)

    def test_render_object_standard_dns_port(self):
        value = '192.0.2.1:53'

        parsed_object = adapters.ZoneMasterAPIv2Adapter.parse_object(
            value, objects.ZoneMaster()
        )

        self.assertEqual('192.0.2.1', parsed_object.host)
        self.assertEqual(53, parsed_object.port)

        render_object = adapters.ZoneMasterAPIv2Adapter.render_object(
            parsed_object
        )

        self.assertEqual('192.0.2.1', render_object)


class ZoneTransferRequestAPIv2AdapterTest(oslotest.base.BaseTestCase):
    def test_parse_object(self):
        values = {
            'target_project_id': '60b1dd98-49a7-403b-901a-291ca555da56',
        }

        parsed_object = adapters.ZoneTransferRequestAPIv2Adapter.parse_object(
            values, objects.ZoneTransferRequest()
        )

        self.assertEqual(
            '60b1dd98-49a7-403b-901a-291ca555da56',
            parsed_object.target_tenant_id
        )

    @mock.patch('designate.policy.check', mock.Mock())
    def test_render_object(self):
        mock_context = mock.Mock()
        values = {
            'target_project_id': '60b1dd98-49a7-403b-901a-291ca555da56',
        }

        parsed_object = adapters.ZoneTransferRequestAPIv2Adapter.parse_object(
            values, objects.ZoneTransferRequest()
        )

        self.assertEqual(
            '60b1dd98-49a7-403b-901a-291ca555da56',
            parsed_object.target_tenant_id
        )

        render_object = adapters.ZoneTransferRequestAPIv2Adapter.render_object(
            parsed_object, context=mock_context
        )

        self.assertEqual(
            '60b1dd98-49a7-403b-901a-291ca555da56',
            render_object['target_project_id']
        )

    @mock.patch('designate.policy.check')
    def test_render_object_forbidden_raised(self, mock_policy_check):
        mock_policy_check.side_effect = exceptions.Forbidden()
        mock_context = mock.Mock()
        values = {
            'target_project_id': '60b1dd98-49a7-403b-901a-291ca555da56',
        }

        parsed_object = adapters.ZoneTransferRequestAPIv2Adapter.parse_object(
            values, objects.ZoneTransferRequest()
        )

        self.assertEqual(
            '60b1dd98-49a7-403b-901a-291ca555da56',
            parsed_object.target_tenant_id
        )

        render_object = adapters.ZoneTransferRequestAPIv2Adapter.render_object(
            parsed_object, context=mock_context
        )

        self.assertNotIn('target_project_id', render_object)



#
# Copyright 2014 Red Hat, Inc.
# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Author: Endre Karlson <endre.karlson@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_config import cfg

COORDINATION_GROUP = cfg.OptGroup(
    name='coordination',
    title="Configuration for coordination"
)

COORDINATION_OPTS = [
    cfg.StrOpt(
        'backend_url',
        secret=True,
        help=(
            'The backend URL to use for distributed coordination. If '
            'unset services that need coordination will function as '
            'a standalone service. This is a `tooz` url - see '
            'https://docs.openstack.org/tooz/latest/user/compatibility.html')
    ),
    cfg.FloatOpt('heartbeat_interval',
                 default=5.0,
                 help='Number of seconds between heartbeats for distributed '
                      'coordination.'),
    cfg.FloatOpt('run_watchers_interval',
                 default=10.0,
                 help='Number of seconds between checks to see if group '
                      'membership has changed'),
]


def register_opts(conf):
    conf.register_group(COORDINATION_GROUP)
    conf.register_opts(COORDINATION_OPTS, group=COORDINATION_GROUP)


def list_opts():
    return {
        COORDINATION_GROUP: COORDINATION_OPTS
    }
